# -*- coding: utf-8 -*-
"""
Created on Tue Nov 04 20:28:06 2014

Lagrange formalism

@author: Topher
"""

import sympy as sp
from sympy import sin,cos,Function

t = sp.Symbol('t')
params = sp.symbols('M , G , J , J_ball , R')
M , G , J , J_ball , R = params

# ball position r
r_t = Function('r')(t)
d_r_t = r_t.diff(t)
dd_r_t = r_t.diff(t,2)
# beam angle theta
theta_t = Function('theta')(t)
d_theta_t = theta_t.diff(t)
dd_theta_t = theta_t.diff(t,2)
# torque of the beam
tau = Function('tau')

# kinetic energy
T = ((M + J_ball/R**2)*d_r_t**2 + (J + M*r_t**2 + J_ball)*d_theta_t**2)/2

# potential energy
V = M*G*r_t*sin(theta_t)

# lagrange function
L = T - V

# replace functions through symbols
r_s , d_r_s , dd_r_s , theta_s, d_theta_s , dd_theta_s = sp.symbols('r_s , d_r_s , dd_r_s , theta_s, d_theta_s , dd_theta_s')
subs_list = [(dd_r_t,dd_r_s),(dd_theta_t,dd_theta_s),(d_r_t,d_r_s),(d_theta_t,d_theta_s),(r_t,r_s),(theta_t,theta_s)]
L = L.subs(subs_list)

# simplify L
L = L.expand()
L = sp.trigsimp(L)

# assitant terms
d_L_r = L.diff(r_s)
d_L_theta = L.diff(theta_s)

d_L_d_r = L.diff(d_r_s)
d_L_d_theta = L.diff(d_theta_s)

# replace symbols through functions
subs_list_rev = []
for tup in subs_list:
    subs_list_rev.append((tup[1],tup[0]))

#d_L_r = d_L_r.subs(subs_list_rev)
#d_L_theta = d_L_theta.subs(subs_list_rev)

d_L_d_r = d_L_d_r.subs(subs_list_rev)
d_L_d_theta = d_L_d_theta.subs(subs_list_rev)

# derivate to time
dt_d_L_d_r = d_L_d_r.diff(t)
dt_d_L_d_theta = d_L_d_theta.diff(t)

# replace functions through symbols
dt_d_L_d_r = dt_d_L_d_r.subs(subs_list)
dt_d_L_d_theta = dt_d_L_d_theta.subs(subs_list)

# lagrange equation
Eq_r = dt_d_L_d_r - d_L_r
Eq_theta = dt_d_L_d_theta - d_L_theta

#Eq_r = Eq_r.factor()
#Eq_theta = Eq_theta.factor()

Eq_r = sp.Eq(0 , Eq_r)
Eq_theta = sp.Eq(tau , Eq_theta)

r_s , d_r_s , dd_r_s , theta_s, d_theta_s , dd_theta_s
# symbols in latex-notation
sn_dict = {r_s: r'r', d_r_s: r'\dot{r}', dd_r_s: r'\ddot{r}',
           theta_s: r' \theta', d_theta_s: r'\dot{\theta}', dd_theta_s: r'\ddot{\theta}'}


# show lagrange equation
import matplotlib.pyplot as plt
latex_str1 = "$ %s $" % sp.latex(Eq_r,symbol_names = sn_dict)
latex_str2 = "$ %s $" % sp.latex(Eq_theta,symbol_names = sn_dict)
latex_str1 = latex_str1.replace("operatorname","mathrm")
latex_str2 = latex_str2.replace("operatorname","mathrm")
plt.text(0.5, 0.5, latex_str1, fontsize=30, horizontalalignment='center')
plt.text(0.5, 0.3, latex_str2, fontsize=30, horizontalalignment='center')
plt.axis('off')
plt.show()


import logging
import os
from abc import ABCMeta, abstractmethod
from pickle import dump

import numpy as np
from scipy.integrate import simps
from PyQt5.QtCore import QObject, QSettings
from PyQt5.QtWidgets import QMessageBox, QFileDialog
pyqtWrapperType = type(QObject)

from .tools import get_sub_value, sort_lists

__all__ = ["PostProcessingModule", "MetaProcessingModule"]

"""
Base Classes for modules in the result-processing environment
"""


class ProcessingModuleMeta(ABCMeta, pyqtWrapperType):
    pass


class ProcessingModule(QObject, metaclass=ProcessingModuleMeta):
    """
    Base Class for processing Modules.

    Each Module's run method is called with a list of results by the
    processing_gui
    """

    # fonts
    _base_font_size = 14
    _title_font_size = 1.5 * _base_font_size
    _label_font_size = 1 * _base_font_size

    # colors
    _grid_color = "#ababab"

    # lines
    _grid_line_style = "--"

    _export_formats = [
        ".pdf",
        # ".png",
        # ".svg",
        # ".eps"
    ]

    def __init__(self, parent=None):
        QObject.__init__(self, parent)
        self.name = self.__class__.__name__
        self._settings = QSettings()
        self._logger = logging.getLogger(self.name)

        self._path_name = None
        self._file_type = None
        self._file_info = None

    @abstractmethod
    def process(self, result_data):
        """
        function that is called when the Processing environment processes all
        loaded result files.

        :param result_data: list of result dicts
        :return: list of diagrams
        """
        pass

    def extract_setting(self, data_list, names, module_name, setting_name):
        """
        extracts settings from simulation data
        :param data_list:
        :param names:
        :param module_name:
        :param setting_name:
        :return:
        """
        return self.extract(data_list, ["modules", module_name, setting_name],
                            names)

    def extract_values(self, data_list, names, value_name):
        """
        extracts values from simulation data
        :param data_list:
        :param names:
        :param value_name:
        :return:
        """
        return self.extract(data_list, ["results", value_name], names)

    def extract(self, data_list, keys, names):
        """
        general extraction from simulation data
        where the regime name contains all strings
        listed in names
        :param data_list:
        :param keys:
        :param names:
        :return:
        """
        if not isinstance(names, list):
            names = [names]

        return next((self._get_sub_dict(result, keys) for result in data_list if
                     all(name in result['regime name'] for name in names)),
                    None)

    @staticmethod
    def _get_sub_dict(top_dict, keys):
        sub_dict = top_dict
        for key in keys:
            sub_dict = sub_dict[key]
        return sub_dict

    def write_output_files(self, result_name, figure, output=None):
        """
        Export the created diagrams and save calculation results in a POF
        (processing output file) file.

        :param result_name:
        :param figure:
        :param output:
        :return:
        """
        path = self._settings.value(self._path_name)
        if not os.path.isdir(path):
            box = QMessageBox()
            box.setText("Export Folder does not exist yet.")
            box.setInformativeText("Do you want to create it? \n"
                                   "{}".format(os.path.abspath(path)))
            box.setStandardButtons(QMessageBox.Ok | QMessageBox.No)
            box.setDefaultButton(QMessageBox.Ok)
            ret = box.exec_()
            if ret == QMessageBox.Ok:
                os.makedirs(path)
            else:
                path = os.path.curdir

        sub_path = os.path.join(path, self.name)
        if not os.path.isdir(sub_path):
            os.makedirs(sub_path)

        dialog = QFileDialog()
        dialog.setAcceptMode(QFileDialog.AcceptSave)
        dialog.setFileMode(QFileDialog.AnyFile)
        dialog.setDirectory(sub_path)
        dialog.setNameFilter("{} (*{})".format(self._file_info,
                                               self._file_type))
        dialog.selectFile(result_name + self._file_type)

        if dialog.exec_():
            file_path = dialog.selectedFiles()[0]
        else:
            self._logger.warning("Export Aborted")
            return

        new_path = os.path.sep.join(file_path.split(os.path.sep)[:-2])
        if new_path != self._settings.value(self._path_name):
            box = QMessageBox()
            box.setText("Use this path as new default?")
            box.setInformativeText("{}".format(new_path))
            box.setStandardButtons(QMessageBox.Yes | QMessageBox.No)
            box.setDefaultButton(QMessageBox.Yes)
            ret = box.exec_()
            if ret == QMessageBox.Yes:
                self._settings.setValue(self._path_name, new_path)

        if output:
            with open(file_path, "wb") as f:
                dump(output, f, protocol=4)

        if figure:
            for export_format in self._export_formats:
                figure.savefig(file_path.split(".")[0] + export_format,
                               bbox_inches='tight')
                # setting bbox_inches='tight' removes the white space around
                # a saved image


class PostProcessingModule(ProcessingModule):
    """
    Base Class for Postprocessing Modules
    """

    def __init__(self):
        ProcessingModule.__init__(self)

        self.figure = None
        self.axes = None
        self.label_counter = 0
        self.results = {}

        self._path_name = "path/postprocessing_results"
        self._file_info = "Processing Output Files"
        self._file_type = ".pof"

    def process(self, files):
        """
        worker-wrapper function that processes an array of result files
        This is an convenience wrapper for simple processor implementation.
        Overload for more sophisticated implementations
        :param files:
        """
        result_list = []
        for data in files:
            self._logger.info("processing data set: {0}"
                              "".format(data["regime name"]))
            result_list.extend(self.run(data))

        return result_list

    @abstractmethod
    def run(self, data):
        """
        Run this postprocessor.
        
        This function will be called from :py:func:`process` with the simulation
        results from one simulation result file.
        
        Overwrite this function to implement your own PostProcessor.

        Args;
        param data: simulation results from a pymoskito simulation result file.
        
        Returns:
             Dict: With a figure Canvas an a name.
        """
        self._logger.warning("placeholder routine called!")
        return {"name": "* placeholder *", "figure": None}

    @staticmethod
    def calc_label_pos(values):
        return np.arange(np.min(values[0]) + 0.1 * values[1],
                         values[1],
                         (values[1] - np.min(values[1])) / 4)

    @staticmethod
    def calc_l1_norm_itae(meas_values, desired_values, step_width):
        """
        Calculate the L1-Norm of the ITAE (Integral of Time-multiplied Absolute
        value of Error).

        Args:
            step_width (float): Time difference between measurements.
            desired_values (array-like): Desired values.
            meas_values (array-like): Measured values.
        """
        def e_func(_t):
            _idx = np.floor_divide(_t, step_width).astype(int)
            e = t * np.abs(desired_values[_idx]
                           - meas_values[_idx])
            return e

        t = np.array([x * step_width for x in range(len(desired_values))])
        err = e_func(t)
        l1norm_itae = simps(err, t)
        return l1norm_itae

    @staticmethod
    def calc_l1_norm_abs(meas_values, desired_values, step_width):
        """
        Calculate the L1-Norm of the absolute error.

        Args:
            step_width (float): Time difference between measurements.
            desired_values (array-like): Desired values.
            meas_values (array-like): Measured values.
        """
        def e_func(_t):
            _idx = np.floor_divide(_t, step_width).astype(int)
            e = np.abs(desired_values[_idx]
                       - meas_values[_idx])
            return e

        t = np.array([x * step_width for x in range(len(desired_values))])
        err = e_func(t)
        l1norm_abs = simps(err, t)
        return l1norm_abs


class MetaProcessingModule(ProcessingModule):
    """
    Base Class for Meta-Processing Modules
    """

    def __init__(self):
        ProcessingModule.__init__(self)

        self._path_name = "path/metaprocessing_results"
        self._file_info = "Meta Output Files"
        self._file_type = ".mof"

    def set_plot_labeling(self, title="",
                          grid=True,
                          x_label="", y_label="",
                          line_type="line"):
        """
        helper to quickly set axis labeling with the good font sizes
        :param title:
        :param grid:
        :param x_label:
        :param y_label:
        :param line_type:
        :return:
        """
        self.axes.set_title(title, size=self._title_font_size)
        self.axes.set_xlabel(x_label, size=self._label_font_size)
        self.axes.set_ylabel(y_label, size=self._label_font_size)

        if grid:
            self.axes.grid(color=self._grid_color,
                           linestyle=self._grid_line_style)

        if line_type != "bar":
            self.axes.legend(loc=0, fontsize='small', prop={'size': 8})

    def plot_family(self, family, x_path, y_path, mode, x_index=-1, y_index=-1):
        """
        plots y over x for all members that can be found in family sources
        :param family:
        :param x_path:
        :param y_path:
        :param mode:
        :param x_index:
        :param y_index:
        :return:
        """
        width = 0.2
        counter = 0
        x_all = []

        for member, data in family.items():
            x_list = get_sub_value(data, x_path)
            y_list = get_sub_value(data, y_path)
            x_list, y_list = sort_lists(x_list, y_list)

            if x_index >= 0:
                x_list[:] = [x[x_index] for x in x_list]
            if y_index >= 0:
                y_list[:] = [y[y_index] for y in y_list]

            # add x values to x_all if there are not in x_all
            for val in x_list:
                if val not in x_all:
                    x_all.append(val)

            if mode == 'line':
                self.axes.plot(x_list, y_list, 'o-', label=member)
            elif mode == 'bar':
                # remove all None from yList
                x_list[:] = [x for x, y in zip(x_list, y_list) if y]
                y_list[:] = [i for i in y_list if i]

                # correction for the position of the bar
                x_list[:] = [k + width * counter for k in x_list]

                self.axes.bar(x_list, y_list, width, label=member)
                counter += 1

        if (mode == 'bar') and (len(x_all) > 1):
            # remove all None from x_all
            x_all.sort()
            x_all[:] = [i for i in x_all if i]

            # does not work for all constellations
            spacing = (x_all[-1] - x_all[0]) / (len(x_all) - 1)
            x_all.append(spacing + x_all[-1])
            x_all.append(x_all[0] - spacing)
            x_all.sort()

            x_all_label = [r'$' + str(i) + '$' for i in x_all]
            counter -= 1
            if mode == 'bar':
                x_all[:] = [i + width * counter for i in x_all]

            self.axes.set_xticks(x_all)
            self.axes.set_xticklabels(x_all_label)



#!/usr/bin/python2
# -*- coding: utf-8 -*-

import numpy as np
import os

import sys
sys.path.append(os.path.pardir)

import settings as st
#-------------------------------------------------------------
# settings
#-------------------------------------------------------------
scale = 100 #in % of param ideal value

controllerList = ['FController', 'GController', 'JController',\
                'LSSController', 'PIFeedbackController']

#-------------------------------------------------------------
# init
#-------------------------------------------------------------
print('\n ### Regimefile Generator Eval B ### \n')

print('Choose parameter to test: ')
for idx, param in enumerate(st.paramVariationListB):
    print('\t', idx, ' - ', param)

paramIdx = -1
while paramIdx not in list(range(0, len(st.paramVariationListB))):
    paramIdx = eval(input())

parameter = st.paramVariationListB[paramIdx]
tol = st.paramToleranceList[paramIdx]

paramRealValue = getattr(st, parameter)
paramTolMinValue = paramRealValue - paramRealValue*tol
paramTolMaxValue = paramRealValue + paramRealValue*tol
paramMinValue = paramRealValue - paramRealValue*tol

print('Choose controller: ')
for idx, controller in enumerate(controllerList):
    print('\t', idx, ' - ', controller)

controlIdx = -1
while controlIdx not in list(range(0, len(controllerList))):
    controlIdx = eval(input())

controller = controllerList[controlIdx]

# how many poles?
if controller == 'PIFeedbackController':
    multiplicator = 5
else:
    multiplicator = 4

#set correct poles
pole = st.poles[controller]

#load head file
filePath = os.path.join(os.curdir, 'B_head.sray')
with open(filePath, 'r') as f:
    head = f.read()

lines = '\n'

#-------------------------------------------------------------
# functions
#-------------------------------------------------------------

#prefined strings
def writeRegime(cName, pole, param, paramVal, appendix):
    return '- name: B_' + cName + '_poles(' + str(pole) + ')'\
                    + '_' + parameter +'(' + str(paramVal) + ')'\
                    + '_' + appendix + '\n'\
                    + '  clear previous: !!python/bool False \n\n'

def writeModel(param, paramVal):
    return '  model:' + '\n'\
            + '   type: BallBeamModel' + '\n'\
            + '   ' + parameter + ': ' + str(paramVal) + '\n\n'

def writeController(cName, pole, multiplicator):
    tmp = '  controller: ' + '\n'\
            + '   type: '  + controller + '\n'\
            + '   poles: ' + str([pole]*multiplicator) + '\n'

    if (controller == 'PIFeedbackController' or controller == 'LSSController'):
        tmp += '   r0: 3'
    tmp += '\n'
    return tmp

#-------------------------------------------------------------
# main
#-------------------------------------------------------------
#ideal regime
lines += writeRegime(controller, pole, parameter, paramRealValue, 'ideal')
lines += writeModel(parameter, paramRealValue)
lines += writeController(controller, pole, multiplicator)

#tolerance bounds
lines += writeRegime(controller, pole, parameter, paramTolMinValue, 'paramTolMin')
lines += writeModel(parameter, paramTolMinValue)
lines += writeController(controller, pole, multiplicator)
lines += writeRegime(controller, pole, parameter, paramTolMaxValue, 'paramTolMax')
lines += writeModel(parameter, paramTolMaxValue)
lines += writeController(controller, pole, multiplicator)

paramBounds = st.paramStabilityLimits[controller][parameter]
simLimits = np.arange(paramBounds[0], paramBounds[1], paramRealValue/scale)

#iterate between limits
#for val in simLimits:
    #lines += writeRegime(controller, pole, parameter, val, 'paramAbs')
    #lines += writeModel(parameter, val)
    #lines += writeController(controller, pole, multiplicator)

dirPath = os.path.join(os.path.pardir, os.path.pardir, 'regimes', 'generated')
if not os.path.isdir(dirPath):
    os.makedirs(dirPath)

fileName = 'B_' + controller + '_poles' + str(pole) \
                    + '_' + str(parameter)\
                    + '(' + str(simLimits[0]) + ',' + str(simLimits[-1])\
                    + ').sreg'

filePath = os.path.join(dirPath, fileName)
with open(filePath, 'w') as f:
    f.write(head)    
    f.write(lines)

print('done.')



# -*- coding: utf-8 -*-
import numpy as np

import pymoskito as pm

import matplotlib as mpl
import matplotlib.patches
import matplotlib.transforms

from . import settings as st

HKS41K100 = '#0b2a51'
HKS44K100 = '#0059a3'
HKS44K80 = '#346FB2'
HKS36K100 = '#512947'
HKS33K100 = '#811a78'
HKS57K100 = '#007a47'
HKS65K100 = '#22ad36'
HKS07K100 = '#e87b14'
HKS07K80 = '#ef9c51'


try:
    import vtk

    class BallBeamVisualizer(pm.VtkVisualizer):

        def __init__(self, renderer):
            pm.VtkVisualizer.__init__(self, renderer)

            # -------- add the beam ----
            # geometry
            self.beam = vtk.vtkCubeSource()
            self.beam.SetXLength(st.visBeamLength)
            self.beam.SetYLength(st.visBeamWidth)
            self.beam.SetZLength(st.visBeamDepth)

            # mapper
            self.beamMapper = vtk.vtkPolyDataMapper()
            self.beamMapper.SetInputConnection(self.beam.GetOutputPort())

            # actor
            self.beamActor = vtk.vtkLODActor()
            self.beamActor.SetMapper(self.beamMapper)

            # make it look nice
            self.beamProp = self.beamActor.GetProperty()
            self.beamProp.SetColor(101 / 255, 123 / 255, 131 / 255)

            self.ren.AddActor(self.beamActor)

            # -------- add the ball ----
            # geometry
            self.ball = vtk.vtkSphereSource()
            self.ball.SetRadius(st.visR)
            self.ball.SetThetaResolution(20)
            self.ball.SetPhiResolution(20)

            # mapper
            self.ballMapper = vtk.vtkPolyDataMapper()
            self.ballMapper.SetInputConnection(self.ball.GetOutputPort())

            # actor
            self.ballActor = vtk.vtkLODActor()
            self.ballActor.SetMapper(self.ballMapper)

            # make it look nice
            self.ballProp = self.ballActor.GetProperty()
            self.ballProp.SetColor(255 / 255, 255 / 255, 0)
            self.ballProp.SetAmbient(0.2)
            self.ballProp.SetDiffuse(0.8)
            self.ballProp.SetSpecular(0.5)
            self.ballProp.SetSpecularPower(0.5)

            self.ren.AddActor(self.ballActor)

            # add background
            self.ren.GradientBackgroundOn()
            self.ren.SetBackground(228 / 255, 232 / 255, 213 / 255)
            self.ren.SetBackground2(38 / 255, 139 / 255, 210 / 255)

            # apply some sane initial state
            self.update_scene(np.array([0, 0, 0, 0]))

            # get everybody into the frame
            self.ren.ResetCamera()
            self.ren.GetActiveCamera().Zoom(1.7)

            # save this view
            self.save_camera_pose()

        def update_config(self, config):
            super().update_config(config)

            model = self._config["Model"]
            radius = st.ballScale * model["R"]
            length = model["beam length"]
            width = model["beam width"]
            depth = model["beam depth"]

            self.ball.SetRadius(radius)
            self.beam.SetXLength(length)
            self.beam.SetYLength(width)
            self.beam.SetZLength(depth)

        def calc_positions(self, x):
            """
            Calculate stationary vectors and rot. matrices for bodies
            """
            model = self._config.get("Model", {})
            radius = st.ballScale * model.get("R", st.R)
            width = model.get("beam width", st.beam_width)

            # beam
            t_beam = np.array([[np.cos(x[2]), -np.sin(x[2]), 0],
                               [np.sin(x[2]), np.cos(x[2]), 0],
                               [0, 0, 1]])
            r_beam0 = np.array([0, -radius - width / 2, 0])
            r_beam = np.dot(t_beam, r_beam0)

            # ball
            r_ball0 = np.array([x[0], 0, 0])
            r_ball = np.dot(t_beam, r_ball0)
            phi = x[0] / radius
            t_ball = np.array([[np.cos(phi), -np.sin(phi), 0],
                               [np.sin(phi), np.cos(phi), 0],
                               [0, 0, 1]])

            return [r_beam, t_beam, r_ball, t_ball]

        @staticmethod
        def set_body_state(actor, r, t):
            poke = vtk.vtkMatrix4x4()

            for i in range(3):
                for n in range(3):
                    poke.SetElement(i, n, t[i, n])
                poke.SetElement(i, 3, r[i])

            actor.PokeMatrix(poke)

        def update_scene(self, x):
            """
            update the body states
            """
            r_beam, t_beam, r_ball, t_ball = self.calc_positions(x)
            self.set_body_state(self.beamActor, r_beam, t_beam)
            self.set_body_state(self.ballActor, r_ball, t_ball)

    pm.register_visualizer(BallBeamVisualizer)

except ImportError as e:
    vtk = None
    print("BallBeam Visualizer:")
    print(e)
    print("VTK Visualization not available.")


class MplBallBeamVisualizer(pm.MplVisualizer):

    def __init__(self, q_widget, q_layout):
        pm.MplVisualizer.__init__(self, q_widget, q_layout)
        self.axes.set_xlim(st.x_min_plot, st.x_max_plot)
        self.axes.set_ylim(st.y_min_plot, st.y_max_plot)
        self.axes.set_aspect("equal")

        self.ball_base = mpl.patches.Circle(
            xy=[0, 0],
            radius=st.visR,
            color=HKS44K100,
            zorder=1)
        self.ball_highlight = mpl.patches.Wedge(
            center=[0, 0],
            r=st.visR,
            theta1=0,
            theta2=90,
            color=HKS07K100,
            zorder=2)
        self.beam = mpl.patches.Rectangle(
            xy=[-st.visBeamLength/2, -(st.visR + st.visBeamWidth)],
            width=st.visBeamLength,
            height=st.visBeamWidth,
            color=HKS41K100,
            zorder=0)

        self.axes.add_patch(self.ball_base)
        self.axes.add_patch(self.ball_highlight)
        self.axes.add_patch(self.beam)

    def update_config(self, config):
        super().update_config(config)

        model = self._config["Model"]
        radius = st.ballScale * model["R"]
        length = model["beam length"]
        width = model["beam width"]

        self.ball_base.set_radius(radius)
        self.ball_highlight.set_radius(radius)
        self.beam.set_xy([-length / 2, -(radius + width)])
        self.beam.set_width(length)
        self.beam.set_height(width)
        self.canvas.draw()

    def update_scene(self, x):
        x_ball, dx_ball, theta_beam, dtheta_beam = x
        theta_ball = -x_ball / st.visR

        t_beam = (mpl.transforms.Affine2D().rotate_around(0, 0, theta_beam)
                  + self.axes.transData)
        t_ball = (mpl.transforms.Affine2D().rotate_around(0, 0, theta_ball)
                  + mpl.transforms.Affine2D().translate(x_ball, 0)
                  + t_beam)

        # ball
        self.ball_base.set_transform(t_ball)
        self.ball_highlight.set_transform(t_ball)

        # beam
        self.beam.set_transform(t_beam)

        self.canvas.draw()

pm.register_visualizer(MplBallBeamVisualizer)



import logging
from abc import ABCMeta, abstractmethod
from collections import OrderedDict
from copy import copy

from PyQt5.QtCore import QObject

pyqtWrapperType = type(QObject)

__all__ = ["SimulationModule", "SimulationException",
           "Trajectory", "Feedforward", "Controller", "Limiter",
           "ModelMixer", "Model", "ModelException",
           "Solver", "Disturbance", "Sensor", "ObserverMixer", "Observer"]


class SimulationModuleMeta(ABCMeta, pyqtWrapperType):
    pass


class SimulationException(Exception):
    pass


class SimulationModule(QObject, metaclass=SimulationModuleMeta):
    """
    Smallest unit pof the simulation framework.

    This class provides necessary functions like output calculation and holds
    all settings that can be accessed by the user.
    The :py:attr:`public_settings` are read by the
    :py:class:`.SimulationInterface` and the rendered by the GUI. All entries
    stated in this dictionary will be available as changeable settings for the
    module.
    On initialization, a possibly modified (in terms of its values) version of
    this dict will be passed back to this class and is thenceforward available
    via the :py:attr:`settings` property.

    The most important method is :py:func:`calc_output` which is called by the
    :py:class:`Simulator` to retrieve this modules output.

    Args:
        settings(OrderedDict): Settings for this simulation module.
            These entries will be shown in the properties view and can be
            changed by the user. The important entries for this base class are:

            `output info`:
                Dict holding an information dictionaries with keys `Name` and
                `Unit` for each element in the output data.
                If available, these information are used to display reasonable
                names in the result view and to display the corresponding units
                for the result plots.

    Warn:
        Do NOT use '.' in the `output_info` name field.

    TODO:
        Get rif of the point restriction
    """

    def __init__(self, settings):
        QObject.__init__(self, None)
        self._logger = logging.getLogger(self.__class__.__name__)

        assert isinstance(settings, dict)
        self._settings = copy(settings)

        self._settings["tick divider"] = settings.get("tick divider", 1)
        self._settings["step width"] = None
        self._settings.pop("modules", None)

    @property
    @abstractmethod
    def public_settings(self):
        pass

    @property
    def settings(self):
        return self._settings

    @property
    def tick_divider(self):
        return self._settings["tick divider"]

    @property
    def step_width(self):
        return self._settings["step width"]

    @step_width.setter
    def step_width(self, value):
        self._settings["step width"] = value

    @abstractmethod
    def calc_output(self, input_vector):
        pass


class ModelException(SimulationException):
    """
    Exception to be raised if the current system state violates modelling
    assumptions.
    """
    pass


class Model(SimulationModule):
    """
    Base class for all user defined system models in state-space form.

    Args:
        settings (dict): Dictionary holding the config options for this module.
            It must contain the following keys:

            :input_count:
                The length of the input vector for this model.

            :state_count:
                The length of the state vector for this model.

            :initial state:
                The initial state vector for this model.
    """

    def __init__(self, settings):
        SimulationModule.__init__(self, settings)
        assert ("state_count" in settings)
        assert ("input_count" in settings)
        assert ("initial state" in settings)
        assert len(settings["initial state"]) == settings["state_count"]

    @property
    def initial_state(self):
        """ Return the initial state of the system. """
        return self._settings["initial state"]

    @abstractmethod
    def state_function(self, t, x, args):
        """
        Calculate the state derivatives of a system with state x at time t.

        Args:
            x(Array-like): System state.
            t(float): System time.
            args: Extra arguments.

        Returns:
            Temporal derivative of the system state at time t.
        """
        pass

    def root_function(self, x):
        """
        Check whether a reinitialisation of the integrator should be performed.
        
        
        This can be the case if there are discontinuities in the system dynamics
        such as switching.
        
        Args:
            x(array-like): Current system state.
            
        Returns: 
            tuple: 
                * bool: `True` if reset is advised.
                * array-like: State to continue with.
                
        """
        return False, x

    def check_consistency(self, x):
        """
        Check whether the assumptions, made in the modelling process are 
        violated.
        
        Args:
            x: Current system state
            
        Raises: 
            :py:class:`ModelException` : If a violation is detected. This will
                stop the simulation process.
        """
        pass


class SolverException(SimulationException):
    pass


class Solver(SimulationModule):
    """
    Base Class for solver implementations
    """

    def __init__(self, settings):
        assert isinstance(settings["modules"]["Model"], Model)
        self._model = settings["modules"]["Model"]
        self.next_output = None
        SimulationModule.__init__(self, settings)

    def calc_output(self, input_vector):
        self.set_input(input_vector["system_input"])
        output = self.next_output

        self.next_output = self.integrate(input_vector["time"])
        try:
            self._model.check_consistency(self.next_output)
        except ModelException as e:
            raise SolverException("Time step integration failed! "
                                  "Model raised: {0}".format(e))
        return output

    @abstractmethod
    def set_input(self, *args):
        pass

    @abstractmethod
    def integrate(self, t):
        pass

    @property
    @abstractmethod
    def t(self):
        pass

    @property
    @abstractmethod
    def successful(self):
        pass


class ControllerException(SimulationException):
    pass


class Controller(SimulationModule):
    """
    Base class for controllers.

    Args:
        settings (dict): Dictionary holding the config options for this module.
            It must contain the following keys:

            :input_order:
                The order of required derivatives from the trajectory generator.

            :input_type:
                Source for the feedback calculation and one of the following:
                `system_state` , `system_output` , `Observer` or `Sensor` .
    """
    # selectable input sources for controller
    input_sources = ["system_state", "system_output", "Observer", "Sensor"]

    def __init__(self, settings):
        SimulationModule.__init__(self, settings)
        assert ("input_order" in settings)
        assert ("input_type" in settings)
        assert (settings["input_type"] in self.input_sources)

    @property
    def input_order(self):
        return self._settings["input_order"]

    def calc_output(self, input_vector):
        input_values = next((input_vector[src] for src in self.input_sources
                             if src == self._settings["input_type"]), None)
        if input_values is None:
            raise ControllerException("Selected Input not available")
        trajectory_values = input_vector.get("Trajectory", None)
        feedforward_values = input_vector.get("Feedforward", None)

        return self._control(input_vector["time"], trajectory_values,
                             feedforward_values, input_values)

    @abstractmethod
    def _control(self, time, trajectory_values=None, feedforward_values=None,
                 input_values=None, **kwargs):
        """
        Placeholder for control law calculations.

        For more sophisticated implementations overload :py:func:`calc_output` .

        Args:
            time (float): Current time.
            trajectory_values (array-like): Desired values from the trajectory
                generator.
            feedforward_values (array-like): Output of feedforward block.
            input_values (array-like): The input values selected by
                ``input_type`` .
            **kwargs: Placeholder for custom parameters.

        Returns:
            Array: Control output.
        """
        pass


class Observer(SimulationModule):
    """
    Base class for observers
    """

    def __init__(self, settings):
        SimulationModule.__init__(self, settings)

    def calc_output(self, input_vector):
        system_input = input_vector.get("system_input", None)
        if "ObserverMixer" in input_vector:
            system_output = input_vector["ObserverMixer"]
        elif "system_output" in input_vector:
            system_output = input_vector["system_output"]
        else:
            raise SimulationException("No Observer input specified")

        return self._observe(input_vector["time"], system_input, system_output)

    @abstractmethod
    def _observe(self, time, system_input, system_output):
        """
        Placeholder for observer law.
        
        Args:
            time: Current time.
            system_input: Current system input.
            system_output: Current system output.

        Returns:
            Estimated system state
        """
        pass


class Feedforward(SimulationModule):
    """
    Base class for all feedforward implementations
    """

    def __init__(self, settings):
        self._model = settings["modules"]["Model"]
        SimulationModule.__init__(self, settings)
        assert ("input_order" in settings)

    @property
    def input_order(self):
        return self._settings["input_order"]

    def calc_output(self, input_dict):
        return self._feedforward(input_dict["time"], input_dict["Trajectory"])

    @abstractmethod
    def _feedforward(self, time, trajectory_values):
        """
        Placeholder for feedforward calculations.

        Args:
            time (float): Current time.
            trajectory_values(array-like): Desired values from the trajectory
                generator.

        Returns:
            Array: Feedforward output. This signal can be added to the
            controllers output via the :py:class:`.ModelMixer` and is also
            directly passed to the controller.
        """
        pass


class TrajectoryException(SimulationException):
    pass


class Trajectory(SimulationModule):
    """
    Base class for all trajectory generators
    """

    def __init__(self, settings):
        if "modules" in settings:
            control_order = 0
            feedforward_order = 0
            if "Controller" in settings["modules"].keys():
                control_order = settings["modules"]["Controller"].input_order
            if "Feedforward" in settings["modules"].keys():
                feedforward_order = settings["modules"]["Feedforward"
                ].input_order
            settings.update(differential_order=max([control_order,
                                                    feedforward_order]))
        else:
            assert "differential_order" in settings
        SimulationModule.__init__(self, settings)

    def calc_output(self, input_vector):
        desired = self._desired_values(input_vector["time"])
        return desired

    @abstractmethod
    def _desired_values(self, t):
        """
        Placeholder for calculations of desired values.

        Args:
            t (float): Time.

        Returns:
            Array: Trajectory output. This should always be a two-dimensional
            array holding the components in to 0th and their derivatives in
            the 1th axis.
        """
        pass


class MixerException(Exception):
    pass


class SignalMixer(SimulationModule):
    """
    Base class for all Signal mixing modules
    """

    def __init__(self, settings):
        assert "input signals" in settings
        SimulationModule.__init__(self, settings)

    def calc_output(self, input_vector):
        signals = [value for signal, value in input_vector.items()
                   if signal in self._settings["input signals"]]
        return self._mix(signals)


class ModelMixer(SignalMixer):
    pass


class ObserverMixer(SignalMixer):
    pass


class Limiter(SimulationModule):
    """
    Base class for all limiter variants
    """

    def __init__(self, settings):
        assert "input_signal" in settings
        SimulationModule.__init__(self, settings)

    def calc_output(self, input_dict):
        return self._limit(input_dict[self._settings["input_signal"]])

    def _limit(self, values):
        """
        Placeholder for actual limit calculations.

        Args:
            values(array-like): Values to limit.

        Returns:
            Array: Limited output.
        """
        return values


class Sensor(SimulationModule):
    """
    Base class for all sensor variants
    """

    def __init__(self, settings):
        assert "input signal" in settings
        SimulationModule.__init__(self, settings)

    def calc_output(self, input_dict):
        return self._measure(input_dict[self._settings["input signal"]])

    def _measure(self, value):
        """
        Placeholder for measurement calculations.

        One may reorder or remove state elements or introduce measurement delays
        here.

        Args:
            value (array-like float): Values from the source selected by the
                ``input_signal`` property.
        Returns:
            array-like float: 'Measured' values.
        """
        return value


class Disturbance(SimulationModule):
    """
    Base class for all disturbance variants
    """

    def __init__(self, settings):
        assert "input signal" in settings
        SimulationModule.__init__(self, settings)

    def calc_output(self, input_dict):
        return self._disturb(input_dict[self._settings["input signal"]])

    @abstractmethod
    def _disturb(self, value):
        """
        Placeholder for disturbance calculations.

        If the noise is to be dependent on the measured signal use its `value`
        to create the noise.

        Args:
            value (array-like float): Values from the source selected by the
                ``input_signal`` property.
        Returns:
            array-like float: Noise that will be mixed with a signal later on.
        """
        pass



# -*- coding: utf-8 -*-
import numpy as np
import pymoskito as pm
import vtk


class TwoPendulumVisualizer(pm.VtkVisualizer):

    def __init__(self, renderer):
        pm.VtkVisualizer.__init__(self, renderer)

        # -------- add the beam ----
        # geometry
        self.beam = vtk.vtkCubeSource()
        self.beam.SetXLength(self.beam_length)
        self.beam.SetYLength(self.beam_height)
        self.beam.SetZLength(self.beam_depth)

        # mapper
        self.beam_Mapper = vtk.vtkPolyDataMapper()
        self.beam_Mapper.SetInputConnection(self.beam.GetOutputPort())

        # actor
        self.beam_Actor = vtk.vtkLODActor()
        self.beam_Actor.SetMapper(self.beam_Mapper)
        self.beam_Actor.SetPosition(0, -(self.cart_height / 2
                                         + self.beam_height / 2), 0)

        # make it look nice
        self.beam_Prop = self.beam_Actor.GetProperty()
        # RAL 9007, grey-aluminium
        self.beam_Prop.SetColor(142 / 255, 142 / 255, 139 / 255)

        self.ren.AddActor(self.beam_Actor)

        # -------- add cart ----
        # geometry
        self.cart = vtk.vtkCubeSource()
        self.cart.SetXLength(self.cart_length)
        self.cart.SetYLength(self.cart_height)
        self.cart.SetZLength(self.cart_depth)

        # mapper
        self.cart_Mapper = vtk.vtkPolyDataMapper()
        self.cart_Mapper.SetInputConnection(self.cart.GetOutputPort())

        # actor
        self.cart_Actor = vtk.vtkLODActor()
        self.cart_Actor.SetPosition(0, 0, 0)
        self.cart_Actor.RotateWXYZ(0, 1, 0, 0)
        self.cart_Actor.SetMapper(self.cart_Mapper)

        # make it look nice
        self.cart_Prop = self.cart_Actor.GetProperty()
        # RAL 7011, iron-grey
        self.cart_Prop.SetColor(64 / 255, 74 / 255, 84 / 255)

        self.ren.AddActor(self.cart_Actor)
        # save pose
        self.cart_pose = [np.array([0, 0, 0]), np.eye(3)]

        # -------- add axis ----
        # geometry
        self.axis = vtk.vtkCylinderSource()
        self.axis.SetHeight(self.axis_height)
        self.axis.SetRadius(self.axis_radius)
        self.axis.SetResolution(100)

        # mapper
        self.axis_Mapper = vtk.vtkPolyDataMapper()
        self.axis_Mapper.SetInputConnection(self.axis.GetOutputPort())

        # actor
        self.axis_Actor = vtk.vtkLODActor()
        self.axis_Actor.SetPosition(0, 0, self.cart_depth/2
                                    + self.axis_height/2)
        self.axis_Actor.RotateWXYZ(90, 1, 0, 0)
        self.axis_Actor.SetMapper(self.axis_Mapper)

        # make it look nice
        self.axis_Prop = self.axis_Actor.GetProperty()
        # RAL 7011, iron-grey
        self.axis_Prop.SetColor(64 / 255, 74 / 255, 84 / 255)

        self.ren.AddActor(self.axis_Actor)
        # save pose
        self.axis_pose = [np.array([0,
                                    0,
                                    self.cart_depth/2 + self.axis_height/2]),
                          pm.rotation_matrix_xyz("x", 90, "deg")]

        # -------- add short pendulum ----
        # add short pendulum shaft
        # geometry
        self.short_pendulum_shaft = vtk.vtkCylinderSource()
        self.short_pendulum_shaft.SetHeight(self.pendulum_shaft_height)
        self.short_pendulum_shaft.SetRadius(self.pendulum_shaft_radius)
        self.short_pendulum_shaft.SetResolution(100)

        # mapper
        self.short_pendulum_shaft_Mapper = vtk.vtkPolyDataMapper()
        self.short_pendulum_shaft_Mapper.SetInputConnection(
            self.short_pendulum_shaft.GetOutputPort())

        # actor
        self.short_pendulum_shaft_Actor = vtk.vtkLODActor()
        self.short_pendulum_shaft_Actor.SetPosition(
            0,
            0,
            self.cart_depth/2
            + self.axis_height
            + self.pendulum_shaft_height/2)
        self.short_pendulum_shaft_Actor.RotateWXYZ(90, 1, 0, 0)
        self.short_pendulum_shaft_Actor.SetMapper(
            self.short_pendulum_shaft_Mapper)

        # make it look nice
        self.short_pendulum_shaft_Prop = \
            self.short_pendulum_shaft_Actor.GetProperty()
        # RAL 9007, grey-aluminium, a little bit darker
        self.short_pendulum_shaft_Prop.SetColor(122 / 255,
                                                122 / 255,
                                                119 / 255)

        self.ren.AddActor(self.short_pendulum_shaft_Actor)
        # save pose
        self.short_pendulum_shaft_pose = [
            np.array([0,
                      0,
                      self.cart_depth/2
                      + self.axis_height
                      + self.pendulum_shaft_height/2]),
            pm.rotation_matrix_xyz("x", 90, "deg")]

        # add short pendulum
        # geometry
        self.short_pendulum = vtk.vtkCylinderSource()
        self.short_pendulum.SetHeight(self.short_pendulum_height)
        self.short_pendulum.SetRadius(self.short_pendulum_radius)
        self.short_pendulum.SetResolution(100)

        # mapper
        self.short_pendulum_Mapper = vtk.vtkPolyDataMapper()
        self.short_pendulum_Mapper.SetInputConnection(
            self.short_pendulum.GetOutputPort())

        # actor
        self.short_pendulum_Actor = vtk.vtkLODActor()
        self.short_pendulum_Actor.SetPosition(0,
                                              self.short_pendulum_height/2,
                                              self.cart_depth/2
                                              + self.axis_height
                                              + self.pendulum_shaft_height/2)
        self.short_pendulum_Actor.RotateWXYZ(0, 1, 0, 0)
        self.short_pendulum_Actor.SetMapper(self.short_pendulum_Mapper)

        # make it look nice
        self.short_pendulum_Prop = self.short_pendulum_Actor.GetProperty()
        # RAL 9007, grey-aluminium, a little bit darker
        self.short_pendulum_Prop.SetColor(122 / 255, 122 / 255, 119 / 255)

        self.ren.AddActor(self.short_pendulum_Actor)
        self.short_pendulum_pose = [
            np.array([0,
                      self.short_pendulum_height/2,
                      self.cart_depth/2
                      + self.axis_height
                      + self.pendulum_shaft_height/2]),
            np.eye(3)]

        # add short pendulum weight
        # geometry
        self.short_pendulum_weight = vtk.vtkCylinderSource()
        self.short_pendulum_weight.SetHeight(self.pendulum_weight_height)
        self.short_pendulum_weight.SetRadius(self.pendulum_weight_radius)
        self.short_pendulum_weight.SetResolution(100)

        # mapper
        self.short_pendulum_weight_Mapper = vtk.vtkPolyDataMapper()
        self.short_pendulum_weight_Mapper.SetInputConnection(
            self.short_pendulum_weight.GetOutputPort())

        # actor
        self.short_pendulum_weight_Actor = vtk.vtkLODActor()
        self.short_pendulum_weight_Actor.SetPosition(
            0,
            (self.short_pendulum_height + self.pendulum_weight_height/2),
            self.cart_depth/2 + self.axis_height + self.pendulum_shaft_height/2)
        self.short_pendulum_weight_Actor.RotateWXYZ(0, 1, 0, 0)
        self.short_pendulum_weight_Actor.SetMapper(
            self.short_pendulum_weight_Mapper)

        # make it look nice
        self.short_pendulum_weight_Prop \
            = self.short_pendulum_weight_Actor.GetProperty()
        # RAL 9006, white-aluminium
        self.short_pendulum_weight_Prop.SetColor(162 / 255,
                                                 166 / 255,
                                                 164 / 255)

        self.ren.AddActor(self.short_pendulum_weight_Actor)
        self.short_pendulum_weight_pose = [
            np.array([0,
                      (self.short_pendulum_height
                       + self.pendulum_weight_height/2),
                      self.cart_depth/2
                      + self.axis_height
                      + self.pendulum_shaft_height/2]),
            np.eye(3)
        ]

        # -------- add long pendulum ----
        # add long pendulum shaft
        # geometry
        self.long_pendulum_shaft = vtk.vtkCylinderSource()
        self.long_pendulum_shaft.SetHeight(self.pendulum_shaft_height)
        self.long_pendulum_shaft.SetRadius(self.pendulum_shaft_radius)
        self.long_pendulum_shaft.SetResolution(100)

        # mapper
        self.long_pendulum_shaft_Mapper = vtk.vtkPolyDataMapper()
        self.long_pendulum_shaft_Mapper.SetInputConnection(
            self.long_pendulum_shaft.GetOutputPort())

        # actor
        self.long_pendulum_shaft_Actor = vtk.vtkLODActor()
        self.long_pendulum_shaft_Actor.SetPosition(
            0,
            0,
            self.cart_depth/2
            + self.axis_height
            + self.pendulum_shaft_height
            + self.pendulum_shaft_height/2)
        self.long_pendulum_shaft_Actor.RotateWXYZ(90, 1, 0, 0)
        self.long_pendulum_shaft_Actor.SetMapper(
            self.long_pendulum_shaft_Mapper)

        # make it look nice
        self.long_pendulum_shaft_Prop = \
            self.long_pendulum_shaft_Actor.GetProperty()
        # RAL 9007, grey-aluminium
        self.long_pendulum_shaft_Prop.SetColor(142 / 255,
                                               142 / 255,
                                               139 / 255)

        self.ren.AddActor(self.long_pendulum_shaft_Actor)
        # save pose
        self.long_pendulum_shaft_pose = [
            np.array([0,
                      0,
                      self.cart_depth/2
                      + self.axis_height
                      + self.pendulum_shaft_height
                      + self.pendulum_shaft_height/2]),
            pm.rotation_matrix_xyz("x", 90, "deg")]

        # add long pendulum
        # geometry
        self.long_pendulum = vtk.vtkCylinderSource()
        self.long_pendulum.SetHeight(self.long_pendulum_height)
        self.long_pendulum.SetRadius(self.long_pendulum_radius)
        self.long_pendulum.SetResolution(100)

        # mapper
        self.long_pendulum_Mapper = vtk.vtkPolyDataMapper()
        self.long_pendulum_Mapper.SetInputConnection(
            self.long_pendulum.GetOutputPort())

        # actor
        self.long_pendulum_Actor = vtk.vtkLODActor()
        self.long_pendulum_Actor.SetPosition(0,
                                             self.long_pendulum_height/2,
                                             self.cart_depth/2
                                             + self.axis_height
                                             + self.pendulum_shaft_height
                                             + self.pendulum_shaft_height/2)
        self.long_pendulum_Actor.RotateWXYZ(0, 1, 0, 0)
        self.long_pendulum_Actor.SetMapper(self.long_pendulum_Mapper)

        # make it look nice
        self.long_pendulum_Prop = self.long_pendulum_Actor.GetProperty()
        # RAL 9007, grey-aluminium
        self.long_pendulum_Prop.SetColor(142 / 255, 142 / 255, 139 / 255)

        self.ren.AddActor(self.long_pendulum_Actor)
        self.long_pendulum_pose = [np.array([0,
                                             self.long_pendulum_height/2,
                                             self.cart_depth/2
                                             + self.axis_height
                                             + self.pendulum_shaft_height
                                             + self.pendulum_shaft_height/2]),
                                   np.eye(3)]

        # add long pendulum weight
        # geometry
        self.long_pendulum_weight = vtk.vtkCylinderSource()
        self.long_pendulum_weight.SetHeight(self.pendulum_weight_height)
        self.long_pendulum_weight.SetRadius(self.pendulum_weight_radius)
        self.long_pendulum_weight.SetResolution(100)

        # mapper
        self.long_pendulum_weight_Mapper = vtk.vtkPolyDataMapper()
        self.long_pendulum_weight_Mapper.SetInputConnection(
            self.long_pendulum_weight.GetOutputPort())

        # actor
        self.long_pendulum_weight_Actor = vtk.vtkLODActor()
        self.long_pendulum_weight_Actor.SetPosition(
            0,
            self.long_pendulum_height + self.pendulum_weight_height/2,
            self.cart_depth/2
            + self.axis_height
            + self.pendulum_shaft_height
            + self.pendulum_shaft_height/2)
        self.long_pendulum_weight_Actor.RotateWXYZ(0, 1, 0, 0)
        self.long_pendulum_weight_Actor.SetMapper(
            self.long_pendulum_weight_Mapper)

        # make it look nice
        self.long_pendulum_weight_Prop \
            = self.long_pendulum_weight_Actor.GetProperty()
        # RAL 9006, white-aluminium
        self.long_pendulum_weight_Prop.SetColor(162 / 255,
                                                166 / 255,
                                                164 / 255)

        self.ren.AddActor(self.long_pendulum_weight_Actor)
        self.long_pendulum_weight_pose = [
            np.array([0,
                      self.long_pendulum_height + self.pendulum_weight_height/2,
                      self.cart_depth/2
                      + self.axis_height
                      + self.pendulum_shaft_height
                      + self.pendulum_shaft_height/2]),
            np.eye(3)]

        # add background
        self.ren.GradientBackgroundOn()
        self.ren.SetBackground(228 / 255, 232 / 255, 213 / 255)
        self.ren.SetBackground2(38 / 255, 139 / 255, 210 / 255)

        # get everybody into the frame
        self.ren.ResetCamera()
        self.ren.GetActiveCamera().Zoom(1.45)
        self.ren.GetActiveCamera().SetFocalPoint(0, 0, 0)

        # save this view
        self.save_camera_pose()

    def calc_positions(self, x):
        """
        Calculate stationary vectors and rot. matrices for bodies
        """
        positions_dict = {}

        # cart
        t_cart = np.array([x[0], 0, 0]) + self.cart_pose[0]
        r_cart = np.eye(3) @ self.cart_pose[1]
        positions_dict.update({"cart": [self.cart_Actor, t_cart, r_cart]})

        # axis
        t_axis = np.array([x[0], 0, 0]) + self.axis_pose[0]
        r_axis = np.eye(3) @ self.axis_pose[1]
        positions_dict.update({"axis": [self.axis_Actor, t_axis, r_axis]})

        # short_pendulum_shaft
        t_short_pendulum_shaft = (np.array([x[0], 0, 0])
                                  + self.short_pendulum_shaft_pose[0])
        r_short_pendulum_shaft = pm.rotation_matrix_xyz(
            "z", x[4], "rad") @ self.short_pendulum_shaft_pose[1]
        positions_dict.update({"short_pendulum_shaft": [
            self.short_pendulum_shaft_Actor,
            t_short_pendulum_shaft,
            r_short_pendulum_shaft]})
        # short_pendulum
        t_short_pendulum = (np.array([x[0], 0, 0])
                            + pm.rotation_matrix_xyz(
            "z", x[4], "rad") @ self.short_pendulum_pose[0])

        r_short_pendulum = pm.rotation_matrix_xyz(
            "z", x[4], "rad") @ self.short_pendulum_pose[1]
        positions_dict.update({"short_pendulum": [self.short_pendulum_Actor,
                                                  t_short_pendulum,
                                                  r_short_pendulum]})

        # short_pendulum_weight
        t_short_pendulum_weight = (np.array([x[0], 0, 0])
                                   + pm.rotation_matrix_xyz(
            "z", x[4], "rad") @ self.short_pendulum_weight_pose[0])
        r_short_pendulum_weight = pm.rotation_matrix_xyz(
            "z", x[4], "rad") @ self.short_pendulum_weight_pose[1]
        positions_dict.update({"short_pendulum_weight": [
            self.short_pendulum_weight_Actor,
            t_short_pendulum_weight,
            r_short_pendulum_weight]})

        # long_pendulum_shaft
        t_long_pendulum_shaft = (np.array([x[0], 0, 0])
                                 + self.long_pendulum_shaft_pose[0])
        r_long_pendulum_shaft = pm.rotation_matrix_xyz(
            "z", x[2], "rad") @ self.long_pendulum_shaft_pose[1]
        positions_dict.update({"long_pendulum_shaft": [
            self.long_pendulum_shaft_Actor,
            t_long_pendulum_shaft,
            r_long_pendulum_shaft]})

        # long_pendulum
        t_long_pendulum = (np.array([x[0], 0, 0])
                           + pm.rotation_matrix_xyz(
            "z", x[2], "rad") @ self.long_pendulum_pose[0])
        r_long_pendulum = pm.rotation_matrix_xyz(
            "z", x[2], "rad") @ self.long_pendulum_pose[1]
        positions_dict.update({"long_pendulum": [self.long_pendulum_Actor,
                                                 t_long_pendulum,
                                                 r_long_pendulum]})

        # long_pendulum_weight
        t_long_pendulum_weight = (np.array([x[0], 0, 0])
                                  + pm.rotation_matrix_xyz(
            "z", x[2], "rad") @ self.long_pendulum_weight_pose[0])
        r_long_pendulum_weight = pm.rotation_matrix_xyz(
            "z", x[2], "rad") @ self.long_pendulum_weight_pose[1]
        positions_dict.update({"long_pendulum_weight": [
            self.long_pendulum_weight_Actor,
            t_long_pendulum_weight,
            r_long_pendulum_weight]})

        return positions_dict

    @staticmethod
    def set_body_state(pos_dict):

        for x in pos_dict.values():
            poke = vtk.vtkMatrix4x4()
            actor = x[0]
            t = x[1]
            r = x[2]
            for i in range(3):
                for n in range(3):
                    poke.SetElement(i, n, r[i, n])
                poke.SetElement(i, 3, t[i])

            actor.PokeMatrix(poke)

    def update_scene(self, x):
        """
        update the body states
        """
        self.set_body_state(self.calc_positions(x))

pm.register_visualizer(TwoPendulumVisualizer)



# -*- coding: utf-8 -*-

# ---------------------------------------------------------------------
# global default settings for physical simulation
# ---------------------------------------------------------------------

# standard values for integration
step_size = 0.002
measure_rate = 500

# initial minimal state vector (n, dn, x, dx)
# default values (used if none is given)
initial_state = [0, 0, 180, 0, 180, 0]

# system values and parameters fo the two pendulum system
g = 9.81                  # m/s**2
# cart:
m0_star = 4.0262          # kg
d0 = 10                   # Ns/m

# long pendulum:
m1_star = 0.3239          # kg
l1_star = 0.446265876476  # m
# d1 = 0.023              # Nms (amira)
d1 = 0.0018468868         # Nms
fv1 = 1.523e-3            # Nms
fc1 = 2.004e-3            # Nm
fh1 = 4.017e-3            # Nm

# short pendulum
m2_star = 0.2357          # kg
l2_star = 0.132090816395  # m
# d2 = 0.00145            # Nms (amira)
d2 = 0.001058521          # Nms
fv2 = 5.974e-9            # Nms
fc2 = 7.717e-3            # Nm
fh2 = 5.451e-3            # Nm

J_DP1 = 0.0723726153903   # kg*m**2
J_DP2 = 0.00428652915258  # kg*m**2
J1_MP = J_DP1 - m1_star*l1_star**2
J2_MP = J_DP2 - m2_star*l2_star**2

# transform the mass and length of the original system in to a
# mass and length of a point-mass-system
l1 = J_DP1/(m1_star*l1_star)
l2 = J_DP2/(m2_star*l2_star)

m1 = (m1_star*l1_star)**2/J_DP1
m2 = (m2_star*l2_star)**2/J_DP2

m0 = m0_star + (m1_star - m1) + (m2_star - m2)

# print m0_star, m0, d0
# print m1_star, m1, l1_star, l1, J_DP1, d1
# print m2_star, m2, l2_star, l2, J_DP2, d2

# information in mm

# beam parameter
beam_length = 1575/1000
beam_height = 10 / 1000
beam_depth = 150/1000

# cart parameter
cart_length = 145/1000
cart_height = 60 / 1000
cart_depth = 135/1000

axis_height = 47.5/1000
axis_radius = (11.8/1000)/2

pendulum_shaft_height = 27/1000
pendulum_shaft_radius = (40/1000)/2

pendulum_weight_height = 40/1000
pendulum_weight_radius = (30/1000)/2

short_pendulum_height = 100/1000 + pendulum_shaft_radius
short_pendulum_radius = (10/1000)/2

long_pendulum_height = 500/1000 + pendulum_shaft_radius
long_pendulum_radius = (10/1000)/2

# xlim for mpl plot
x_min_plot = -850/1000
x_max_plot = 850/1000

y_min_plot = -600/1000
y_max_plot = 600/1000



# -*- coding: utf-8 -*-
import numpy as np
import pymoskito as pm

import matplotlib as mpl
import matplotlib.patches
import matplotlib.transforms

from . import settings as st

try:
    import vtk

    class TwoPendulumVisualizer(pm.VtkVisualizer):

        def __init__(self, renderer):
            pm.VtkVisualizer.__init__(self, renderer)

            # -------- add the beam ----
            # geometry
            self.beam = vtk.vtkCubeSource()
            self.beam.SetXLength(st.beam_length)
            self.beam.SetYLength(st.beam_height)
            self.beam.SetZLength(st.beam_depth)

            # mapper
            self.beam_Mapper = vtk.vtkPolyDataMapper()
            self.beam_Mapper.SetInputConnection(self.beam.GetOutputPort())

            # actor
            self.beam_Actor = vtk.vtkLODActor()
            self.beam_Actor.SetMapper(self.beam_Mapper)
            self.beam_Actor.SetPosition(0, -(st.cart_height / 2
                                             + st.beam_height / 2), 0)

            # make it look nice
            self.beam_Prop = self.beam_Actor.GetProperty()
            # RAL 9007, grey-aluminium
            self.beam_Prop.SetColor(142 / 255, 142 / 255, 139 / 255)

            self.ren.AddActor(self.beam_Actor)

            # -------- add cart ----
            # geometry
            self.cart = vtk.vtkCubeSource()
            self.cart.SetXLength(st.cart_length)
            self.cart.SetYLength(st.cart_height)
            self.cart.SetZLength(st.cart_depth)

            # mapper
            self.cart_Mapper = vtk.vtkPolyDataMapper()
            self.cart_Mapper.SetInputConnection(self.cart.GetOutputPort())

            # actor
            self.cart_Actor = vtk.vtkLODActor()
            self.cart_Actor.SetPosition(0, 0, 0)
            self.cart_Actor.RotateWXYZ(0, 1, 0, 0)
            self.cart_Actor.SetMapper(self.cart_Mapper)

            # make it look nice
            self.cart_Prop = self.cart_Actor.GetProperty()
            # RAL 7011, iron-grey
            self.cart_Prop.SetColor(64 / 255, 74 / 255, 84 / 255)

            self.ren.AddActor(self.cart_Actor)
            # save pose
            self.cart_pose = [np.array([0, 0, 0]), np.eye(3)]

            # -------- add axis ----
            # geometry
            self.axis = vtk.vtkCylinderSource()
            self.axis.SetHeight(st.axis_height)
            self.axis.SetRadius(st.axis_radius)
            self.axis.SetResolution(100)

            # mapper
            self.axis_Mapper = vtk.vtkPolyDataMapper()
            self.axis_Mapper.SetInputConnection(self.axis.GetOutputPort())

            # actor
            self.axis_Actor = vtk.vtkLODActor()
            self.axis_Actor.SetPosition(0, 0, st.cart_depth/2
                                        + st.axis_height/2)
            self.axis_Actor.RotateWXYZ(90, 1, 0, 0)
            self.axis_Actor.SetMapper(self.axis_Mapper)

            # make it look nice
            self.axis_Prop = self.axis_Actor.GetProperty()
            # RAL 7011, iron-grey
            self.axis_Prop.SetColor(64 / 255, 74 / 255, 84 / 255)

            self.ren.AddActor(self.axis_Actor)
            # save pose
            self.axis_pose = [np.array([0,
                                        0,
                                        st.cart_depth/2 + st.axis_height/2]),
                              pm.rotation_matrix_xyz("x", 90, "deg")]

            # -------- add short pendulum ----
            # add short pendulum shaft
            # geometry
            self.short_pendulum_shaft = vtk.vtkCylinderSource()
            self.short_pendulum_shaft.SetHeight(st.pendulum_shaft_height)
            self.short_pendulum_shaft.SetRadius(st.pendulum_shaft_radius)
            self.short_pendulum_shaft.SetResolution(100)

            # mapper
            self.short_pendulum_shaft_Mapper = vtk.vtkPolyDataMapper()
            self.short_pendulum_shaft_Mapper.SetInputConnection(
                self.short_pendulum_shaft.GetOutputPort())

            # actor
            self.short_pendulum_shaft_Actor = vtk.vtkLODActor()
            self.short_pendulum_shaft_Actor.SetPosition(
                0,
                0,
                st.cart_depth/2
                + st.axis_height
                + st.pendulum_shaft_height/2)
            self.short_pendulum_shaft_Actor.RotateWXYZ(90, 1, 0, 0)
            self.short_pendulum_shaft_Actor.SetMapper(
                self.short_pendulum_shaft_Mapper)

            # make it look nice
            self.short_pendulum_shaft_Prop = \
                self.short_pendulum_shaft_Actor.GetProperty()
            # RAL 9007, grey-aluminium, a little bit darker
            self.short_pendulum_shaft_Prop.SetColor(122 / 255,
                                                    122 / 255,
                                                    119 / 255)

            self.ren.AddActor(self.short_pendulum_shaft_Actor)
            # save pose
            self.short_pendulum_shaft_pose = [
                np.array([0,
                          0,
                          st.cart_depth/2
                          + st.axis_height
                          + st.pendulum_shaft_height/2]),
                pm.rotation_matrix_xyz("x", 90, "deg")]

            # add short pendulum
            # geometry
            self.short_pendulum = vtk.vtkCylinderSource()
            self.short_pendulum.SetHeight(st.short_pendulum_height)
            self.short_pendulum.SetRadius(st.short_pendulum_radius)
            self.short_pendulum.SetResolution(100)

            # mapper
            self.short_pendulum_Mapper = vtk.vtkPolyDataMapper()
            self.short_pendulum_Mapper.SetInputConnection(
                self.short_pendulum.GetOutputPort())

            # actor
            self.short_pendulum_Actor = vtk.vtkLODActor()
            self.short_pendulum_Actor.SetPosition(0,
                                                  st.short_pendulum_height/2,
                                                  st.cart_depth/2
                                                  + st.axis_height
                                                  + st.pendulum_shaft_height/2)
            self.short_pendulum_Actor.RotateWXYZ(0, 1, 0, 0)
            self.short_pendulum_Actor.SetMapper(self.short_pendulum_Mapper)

            # make it look nice
            self.short_pendulum_Prop = self.short_pendulum_Actor.GetProperty()
            # RAL 9007, grey-aluminium, a little bit darker
            self.short_pendulum_Prop.SetColor(122 / 255, 122 / 255, 119 / 255)

            self.ren.AddActor(self.short_pendulum_Actor)
            self.short_pendulum_pose = [
                np.array([0,
                          st.short_pendulum_height/2,
                          st.cart_depth/2
                          + st.axis_height
                          + st.pendulum_shaft_height/2]),
                np.eye(3)]

            # add short pendulum weight
            # geometry
            self.short_pendulum_weight = vtk.vtkCylinderSource()
            self.short_pendulum_weight.SetHeight(st.pendulum_weight_height)
            self.short_pendulum_weight.SetRadius(st.pendulum_weight_radius)
            self.short_pendulum_weight.SetResolution(100)

            # mapper
            self.short_pendulum_weight_Mapper = vtk.vtkPolyDataMapper()
            self.short_pendulum_weight_Mapper.SetInputConnection(
                self.short_pendulum_weight.GetOutputPort())

            # actor
            self.short_pendulum_weight_Actor = vtk.vtkLODActor()
            self.short_pendulum_weight_Actor.SetPosition(
                0,
                (st.short_pendulum_height + st.pendulum_weight_height/2),
                st.cart_depth/2 + st.axis_height + st.pendulum_shaft_height/2)
            self.short_pendulum_weight_Actor.RotateWXYZ(0, 1, 0, 0)
            self.short_pendulum_weight_Actor.SetMapper(
                self.short_pendulum_weight_Mapper)

            # make it look nice
            self.short_pendulum_weight_Prop \
                = self.short_pendulum_weight_Actor.GetProperty()
            # RAL 9006, white-aluminium
            self.short_pendulum_weight_Prop.SetColor(162 / 255,
                                                     166 / 255,
                                                     164 / 255)

            self.ren.AddActor(self.short_pendulum_weight_Actor)
            self.short_pendulum_weight_pose = [
                np.array([0,
                          (st.short_pendulum_height
                           + st.pendulum_weight_height/2),
                          st.cart_depth/2
                          + st.axis_height
                          + st.pendulum_shaft_height/2]),
                np.eye(3)
            ]

            # -------- add long pendulum ----
            # add long pendulum shaft
            # geometry
            self.long_pendulum_shaft = vtk.vtkCylinderSource()
            self.long_pendulum_shaft.SetHeight(st.pendulum_shaft_height)
            self.long_pendulum_shaft.SetRadius(st.pendulum_shaft_radius)
            self.long_pendulum_shaft.SetResolution(100)

            # mapper
            self.long_pendulum_shaft_Mapper = vtk.vtkPolyDataMapper()
            self.long_pendulum_shaft_Mapper.SetInputConnection(
                self.long_pendulum_shaft.GetOutputPort())

            # actor
            self.long_pendulum_shaft_Actor = vtk.vtkLODActor()
            self.long_pendulum_shaft_Actor.SetPosition(
                0,
                0,
                st.cart_depth/2
                + st.axis_height
                + st.pendulum_shaft_height
                + st.pendulum_shaft_height/2)
            self.long_pendulum_shaft_Actor.RotateWXYZ(90, 1, 0, 0)
            self.long_pendulum_shaft_Actor.SetMapper(
                self.long_pendulum_shaft_Mapper)

            # make it look nice
            self.long_pendulum_shaft_Prop =\
                self.long_pendulum_shaft_Actor.GetProperty()
            # RAL 9007, grey-aluminium
            self.long_pendulum_shaft_Prop.SetColor(142 / 255,
                                                   142 / 255,
                                                   139 / 255)

            self.ren.AddActor(self.long_pendulum_shaft_Actor)
            # save pose
            self.long_pendulum_shaft_pose = [
                np.array([0,
                          0,
                          st.cart_depth/2
                          + st.axis_height
                          + st.pendulum_shaft_height
                          + st.pendulum_shaft_height/2]),
                pm.rotation_matrix_xyz("x", 90, "deg")]

            # add long pendulum
            # geometry
            self.long_pendulum = vtk.vtkCylinderSource()
            self.long_pendulum.SetHeight(st.long_pendulum_height)
            self.long_pendulum.SetRadius(st.long_pendulum_radius)
            self.long_pendulum.SetResolution(100)

            # mapper
            self.long_pendulum_Mapper = vtk.vtkPolyDataMapper()
            self.long_pendulum_Mapper.SetInputConnection(
                self.long_pendulum.GetOutputPort())

            # actor
            self.long_pendulum_Actor = vtk.vtkLODActor()
            self.long_pendulum_Actor.SetPosition(0,
                                                 st.long_pendulum_height/2,
                                                 st.cart_depth/2
                                                 + st.axis_height
                                                 + st.pendulum_shaft_height
                                                 + st.pendulum_shaft_height/2)
            self.long_pendulum_Actor.RotateWXYZ(0, 1, 0, 0)
            self.long_pendulum_Actor.SetMapper(self.long_pendulum_Mapper)

            # make it look nice
            self.long_pendulum_Prop = self.long_pendulum_Actor.GetProperty()
            # RAL 9007, grey-aluminium
            self.long_pendulum_Prop.SetColor(142 / 255, 142 / 255, 139 / 255)

            self.ren.AddActor(self.long_pendulum_Actor)
            self.long_pendulum_pose = [np.array([0,
                                                 st.long_pendulum_height/2,
                                                 st.cart_depth/2
                                                 + st.axis_height
                                                 + st.pendulum_shaft_height
                                                 + st.pendulum_shaft_height/2]),
                                       np.eye(3)]

            # add long pendulum weight
            # geometry
            self.long_pendulum_weight = vtk.vtkCylinderSource()
            self.long_pendulum_weight.SetHeight(st.pendulum_weight_height)
            self.long_pendulum_weight.SetRadius(st.pendulum_weight_radius)
            self.long_pendulum_weight.SetResolution(100)

            # mapper
            self.long_pendulum_weight_Mapper = vtk.vtkPolyDataMapper()
            self.long_pendulum_weight_Mapper.SetInputConnection(
                self.long_pendulum_weight.GetOutputPort())

            # actor
            self.long_pendulum_weight_Actor = vtk.vtkLODActor()
            self.long_pendulum_weight_Actor.SetPosition(
                0,
                st.long_pendulum_height + st.pendulum_weight_height/2,
                st.cart_depth/2
                + st.axis_height
                + st.pendulum_shaft_height
                + st.pendulum_shaft_height/2)
            self.long_pendulum_weight_Actor.RotateWXYZ(0, 1, 0, 0)
            self.long_pendulum_weight_Actor.SetMapper(
                self.long_pendulum_weight_Mapper)

            # make it look nice
            self.long_pendulum_weight_Prop \
                = self.long_pendulum_weight_Actor.GetProperty()
            # RAL 9006, white-aluminium
            self.long_pendulum_weight_Prop.SetColor(162 / 255,
                                                    166 / 255,
                                                    164 / 255)

            self.ren.AddActor(self.long_pendulum_weight_Actor)
            self.long_pendulum_weight_pose = [
                np.array([0,
                          st.long_pendulum_height + st.pendulum_weight_height/2,
                          st.cart_depth/2
                          + st.axis_height
                          + st.pendulum_shaft_height
                          + st.pendulum_shaft_height/2]),
                np.eye(3)]

            # add background
            self.ren.GradientBackgroundOn()
            self.ren.SetBackground(228 / 255, 232 / 255, 213 / 255)
            self.ren.SetBackground2(38 / 255, 139 / 255, 210 / 255)

            # get everybody into the frame
            self.ren.ResetCamera()
            self.ren.GetActiveCamera().Zoom(1.45)
            self.ren.GetActiveCamera().SetFocalPoint(0, 0, 0)

            # save this view
            self.save_camera_pose()

        def calc_positions(self, x):
            """
            Calculate stationary vectors and rot. matrices for bodies
            """
            positions_dict = {}

            # cart
            t_cart = np.array([x[0], 0, 0]) + self.cart_pose[0]
            r_cart = np.eye(3) @ self.cart_pose[1]
            positions_dict.update({"cart": [self.cart_Actor, t_cart, r_cart]})

            # axis
            t_axis = np.array([x[0], 0, 0]) + self.axis_pose[0]
            r_axis = np.eye(3) @ self.axis_pose[1]
            positions_dict.update({"axis": [self.axis_Actor, t_axis, r_axis]})

            # short_pendulum_shaft
            t_short_pendulum_shaft = (np.array([x[0], 0, 0])
                                      + self.short_pendulum_shaft_pose[0])
            r_short_pendulum_shaft = pm.rotation_matrix_xyz(
                "z", x[4], "rad") @ self.short_pendulum_shaft_pose[1]
            positions_dict.update({"short_pendulum_shaft": [
                self.short_pendulum_shaft_Actor,
                t_short_pendulum_shaft,
                r_short_pendulum_shaft]})
            # short_pendulum
            t_short_pendulum = (np.array([x[0], 0, 0])
                                + pm.rotation_matrix_xyz(
                "z", x[4], "rad") @ self.short_pendulum_pose[0])

            r_short_pendulum = pm.rotation_matrix_xyz(
                "z", x[4], "rad") @ self.short_pendulum_pose[1]
            positions_dict.update({"short_pendulum": [self.short_pendulum_Actor,
                                                      t_short_pendulum,
                                                      r_short_pendulum]})

            # short_pendulum_weight
            t_short_pendulum_weight = (np.array([x[0], 0, 0])
                                       + pm.rotation_matrix_xyz(
                "z", x[4], "rad") @ self.short_pendulum_weight_pose[0])
            r_short_pendulum_weight = pm.rotation_matrix_xyz(
                "z", x[4], "rad") @ self.short_pendulum_weight_pose[1]
            positions_dict.update({"short_pendulum_weight": [
                self.short_pendulum_weight_Actor,
                t_short_pendulum_weight,
                r_short_pendulum_weight]})

            # long_pendulum_shaft
            t_long_pendulum_shaft = (np.array([x[0], 0, 0])
                                     + self.long_pendulum_shaft_pose[0])
            r_long_pendulum_shaft = pm.rotation_matrix_xyz(
                "z", x[2], "rad") @ self.long_pendulum_shaft_pose[1]
            positions_dict.update({"long_pendulum_shaft": [
                self.long_pendulum_shaft_Actor,
                t_long_pendulum_shaft,
                r_long_pendulum_shaft]})

            # long_pendulum
            t_long_pendulum = (np.array([x[0], 0, 0])
                               + pm.rotation_matrix_xyz(
                "z", x[2], "rad") @ self.long_pendulum_pose[0])
            r_long_pendulum = pm.rotation_matrix_xyz(
                "z", x[2], "rad") @ self.long_pendulum_pose[1]
            positions_dict.update({"long_pendulum": [self.long_pendulum_Actor,
                                                     t_long_pendulum,
                                                     r_long_pendulum]})

            # long_pendulum_weight
            t_long_pendulum_weight = (np.array([x[0], 0, 0])
                                      + pm.rotation_matrix_xyz(
                "z", x[2], "rad") @ self.long_pendulum_weight_pose[0])
            r_long_pendulum_weight = pm.rotation_matrix_xyz(
                "z", x[2], "rad") @ self.long_pendulum_weight_pose[1]
            positions_dict.update({"long_pendulum_weight": [
                self.long_pendulum_weight_Actor,
                t_long_pendulum_weight,
                r_long_pendulum_weight]})

            return positions_dict

        @staticmethod
        def set_body_state(pos_dict):

            for x in pos_dict.values():
                poke = vtk.vtkMatrix4x4()
                actor = x[0]
                t = x[1]
                r = x[2]
                for i in range(3):
                    for n in range(3):
                        poke.SetElement(i, n, r[i, n])
                    poke.SetElement(i, 3, t[i])

                actor.PokeMatrix(poke)

        def update_scene(self, x):
            """
            update the body states
            """
            self.set_body_state(self.calc_positions(x))

    pm.register_visualizer(TwoPendulumVisualizer)

except ImportError as e:
    vtk = None
    print("BallTube Visualizer:")
    print(e)
    print("VTK Visualization not available.")


class MplTwoPendulumVisualizer(pm.MplVisualizer):

    def __init__(self, q_widget, q_layout):
        pm.MplVisualizer.__init__(self, q_widget, q_layout)
        self.axes.set_xlim(st.x_min_plot, st.x_max_plot)
        self.axes.set_ylim(st.y_min_plot, st.y_max_plot)
        self.axes.set_aspect("equal")

        self.beam = mpl.patches.Rectangle(xy=[-st.beam_length/2,
                                              -(st.beam_height
                                                + st.cart_height/2)],
                                          width=st.beam_length,
                                          height=st.beam_height,
                                          color="lightgrey")

        self.cart = mpl.patches.Rectangle(xy=[-st.cart_length/2,
                                              -st.cart_height/2],
                                          width=st.cart_length,
                                          height=st.cart_height,
                                          color="dimgrey")

        self.pendulum_shaft = mpl.patches.Circle(
            xy=[0, 0],
            radius=st.pendulum_shaft_radius,
            color="lightgrey",
            zorder=3)

        t = mpl.transforms.Affine2D().rotate_deg(180) + self.axes.transData
        self.short_pendulum = mpl.patches.Rectangle(
            xy=[-st.short_pendulum_radius, 0],
            width=2*st.short_pendulum_radius,
            height=st.short_pendulum_height,
            color="#E87B14",  # TUD CD HKS 07_K
            zorder=2,
            transform=t)
        self.short_pendulum_weight = mpl.patches.Rectangle(
            xy=[-st.pendulum_weight_radius, st.short_pendulum_height],
            width=2*st.pendulum_weight_radius,
            height=st.pendulum_weight_height,
            color="#E87B14",  # TUD CD HKS 07_K
            zorder=2,
            transform=t)
        self.long_pendulum = mpl.patches.Rectangle(
            xy=[-st.long_pendulum_radius, 0],
            width=2*st.long_pendulum_radius,
            height=st.long_pendulum_height,
            color="#0059A3",  # TUD CD HKS 44_K
            zorder=1,
            transform=t)
        self.long_pendulum_weight = mpl.patches.Rectangle(
            xy=[-st.pendulum_weight_radius, st.long_pendulum_height],
            width=2*st.pendulum_weight_radius,
            height=st.pendulum_weight_height,
            color="#0059A3",  # TUD CD HKS 44_K
            zorder=1,
            transform=t)

        self.axes.add_patch(self.beam)
        self.axes.add_patch(self.cart)
        self.axes.add_patch(self.pendulum_shaft)
        self.axes.add_patch(self.short_pendulum)
        self.axes.add_patch(self.short_pendulum_weight)
        self.axes.add_patch(self.long_pendulum)
        self.axes.add_patch(self.long_pendulum_weight)

    def update_scene(self, x):
        x0 = x[0]
        phi1 = np.rad2deg(x[2])
        phi2 = np.rad2deg(x[4])

        # cart and shaft
        self.cart.set_x(-st.cart_length/2 + x0)
        self.pendulum_shaft.center = [x0, 0]

        t_phi1 = (mpl.transforms.Affine2D().rotate_deg_around(x0, 0, phi1)
                  + self.axes.transData)
        t_phi2 = (mpl.transforms.Affine2D().rotate_deg_around(x0, 0, phi2)
                  + self.axes.transData)

        # long pendulum
        self.long_pendulum.set_xy([-st.long_pendulum_radius + x0, 0])
        self.long_pendulum.set_transform(t_phi1)
        self.long_pendulum_weight.set_xy([-st.pendulum_weight_radius
                                          + x0, st.long_pendulum_height])
        self.long_pendulum_weight.set_transform(t_phi1)

        # short pendulum
        self.short_pendulum.set_xy(np.array([-st.short_pendulum_radius, 0])
                                   + np.array([x0, 0]))
        self.short_pendulum.set_transform(t_phi2)
        self.short_pendulum_weight.set_xy([-st.pendulum_weight_radius + x0,
                                           st.short_pendulum_height])

        self.short_pendulum_weight.set_transform(t_phi2)
        self.canvas.draw()

pm.register_visualizer(MplTwoPendulumVisualizer)



# -*- coding: utf-8 -*-
import numpy as np
import matplotlib as mpl
import matplotlib.patches

import pymoskito as pm

from . import settings as st

try:
    import vtk

    class BallInTubeVisualizer(pm.VtkVisualizer):

        def __init__(self, renderer):
            pm.VtkVisualizer.__init__(self, renderer)

            # -------- add a line ----
            # geometry
            self.line = vtk.vtkLineSource()
            self.line.SetPoint1(0, 0, 0)
            self.line.SetPoint2(0, +st.visTubeLength, 0)

            # mapper
            self.lineMapper = vtk.vtkPolyDataMapper()
            self.lineMapper.SetInputConnection(self.line.GetOutputPort())

            # actor
            self.lineActor = vtk.vtkLODActor()
            self.lineActor.SetMapper(self.lineMapper)

            # make it look nice
            self.lineProp = self.lineActor.GetProperty()
            self.lineProp.SetOpacity(0.0)

            self.ren.AddActor(self.lineActor)

            # -------- add a tube around the line ----
            # geometry
            self.tube = vtk.vtkTubeFilter()
            self.tube.SetNumberOfSides(100)
            self.tube.SetInputConnection(self.line.GetOutputPort())
            self.tube.SetRadius(st.visTubeDiameter/2)

            # mapper
            self.tubeMapper = vtk.vtkPolyDataMapper()
            self.tubeMapper.SetInputConnection(self.tube.GetOutputPort())

            # actor
            self.tubeActor = vtk.vtkLODActor()
            self.tubeActor.SetMapper(self.tubeMapper)

            # make it look nice
            self.tubeProp = self.tubeActor.GetProperty()
            self.tubeProp.SetColor(101/255, 123/255, 131/255)
            self.tubeProp.SetOpacity(0.5)

            self.ren.AddActor(self.tubeActor)

            # -------- add the ball ----
            # geometry
            self.ball = vtk.vtkSphereSource()
            self.ball.SetRadius(st.visBallDiameter/2)
            self.ball.SetThetaResolution(30)
            self.ball.SetPhiResolution(30)
            self.ball.SetCenter(0, st.visBallDiameter/2, 0)

            # mapper
            self.ballMapper = vtk.vtkPolyDataMapper()
            self.ballMapper.SetInputConnection(self.ball.GetOutputPort())

            # actor
            self.ballActor = vtk.vtkLODActor()
            self.ballActor.SetMapper(self.ballMapper)

            # make it look nice
            self.ballProp = self.ballActor.GetProperty()
            self.ballProp.SetColor(255/255, 255/255, 0)
            self.ballProp.SetAmbient(0.2)
            self.ballProp.SetDiffuse(0.8)
            self.ballProp.SetSpecular(0.5)
            self.ballProp.SetSpecularPower(0.5)

            self.ren.AddActor(self.ballActor)

            # -------- add background ----
            self.ren.GradientBackgroundOn()
            self.ren.SetBackground(228/255, 232/255, 213/255)
            self.ren.SetBackground2(38/255, 139/255, 210/255)

            # get everybody into the frame
            self.ren.ResetCamera()
            self.ren.GetActiveCamera().Zoom(1.0)

            # save this view
            self.save_camera_pose()

        @staticmethod
        def calc_positions(x):
            """
            Calculate stationary vectors and rot. matrices for bodies
            """
            # not used, because the fan is not visualized
            # r_fan = np.zeros(3)
            # t_fan = np.zeros((3, 3))

            r_ball = np.array([0, x[2]/st.scale, 0])
            t_ball = np.eye(3)

            return [r_ball, t_ball]

        @staticmethod
        def set_body_state(actor, r, t):
            poke = vtk.vtkMatrix4x4()

            for i in range(3):
                for n in range(3):
                    poke.SetElement(i, n, t[i, n])
                poke.SetElement(i, 3, r[i])

            actor.PokeMatrix(poke)

        def update_scene(self, x):
            """
            update the body states
            """
            r_ball, t_ball = self.calc_positions(x)
            self.set_body_state(self.ballActor, r_ball, t_ball)

    pm.register_visualizer(BallInTubeVisualizer)

except ImportError as e:
    vtk = None
    print("BallTube Visualizer:")
    print(e)
    print("VTK Visualization not available.")


class MplBallInTubeVisualizer(pm.MplVisualizer):

    def __init__(self, q_widget, q_layout):
        pm.MplVisualizer.__init__(self, q_widget, q_layout)

        self.axes.set_xlim(-0.3, 0.3)
        self.axes.set_ylim(-0.05, 1.55)
        self.axes.set_aspect("equal")
        self.axes.get_xaxis().set_visible(False)
#        tube = mpl.patches.Rectangle(xy=[-st.d_R*st.scale/2.0, 0],
#                                     width=st.d_R*st.scale,
#                                     height=st.tube_length,
#                                     fill=False)
#        self.axes.add_patch(tube)
        tube_out = mpl.patches.Rectangle(xy=[-st.d_R_out * st.scale / 2.0, 0],
                                         width=st.d_R_out * st.scale,
                                         height=st.tube_length,
                                         linewidth=1,
                                         fill=False)
        self.axes.add_patch(tube_out)

        self.ball = mpl.patches.Circle(xy=[0, st.d_B*st.scale/2.0],
                                       radius=st.d_B*st.scale/2.0,
                                       color="#0059A3",
                                       linewidth=1)
        self.ball.set_edgecolor("black")
        self.axes.add_patch(self.ball)

    def update_scene(self, x):

        self.ball.center = (0, x[2] + st.d_R*st.scale/2.0)
        self.canvas.draw()

pm.register_visualizer(MplBallInTubeVisualizer)



#!/usr/bin/python2
# -*- coding: utf-8 -*-

import numpy as np
import os

import sys
sys.path.append(os.path.pardir)

import settings as st
#-------------------------------------------------------------
# settings
#-------------------------------------------------------------
end_time = 20

controllerList = ['FController', 'GController', 'JController',\
                'LSSController', 'PIFeedbackController', 'all']

#-------------------------------------------------------------
# init
#-------------------------------------------------------------
print('\n ### Regimefile Generator Eval C1 ### \n')
print('Choose parameter to test: ')
parameterList = ['sigma', 'delay']
for idx, param in enumerate(parameterList):
    print('\t', idx, ' - ', param)

paramIdx = -1
while paramIdx not in list(range(0, len(parameterList))):
    paramIdx = eval(input())

# extract parameter values
parameter = parameterList[paramIdx]

print('Choose controller: ')
for idx, controller in enumerate(controllerList):
    print('\t', idx, ' - ', controller)

controlIdx = -1
while controlIdx not in list(range(0, len(controllerList))):
    controlIdx = eval(input())

if controlIdx == len(controllerList)-1:
    simCases = controllerList[:-1]
else:
    simCases = [controllerList[controlIdx]]

#-------------------------------------------------------------
# functions
#-------------------------------------------------------------

#prefined strings
def writeRegime(cName, pole, param, paramVal, appendix):
    return '- name: C1_' + cName + '_poles(' + str(pole) + ')'\
                    + '_' + parameter +'(' + str(paramVal) + ')'\
                    + appendix + '\n'\
                    + '  clear previous: !!python/bool False \n\n'

def writeModel(param, paramVal):
    return '  model:' + '\n'\
            + '   type: BallBeamModel' + '\n'\
            + '   ' + parameter + ': ' + str(paramVal) + '\n\n'

def writeController(cName, pole, multiplicator):
    tmp = '  controller: ' + '\n'\
            + '   type: '  + controller + '\n'\
            + '   poles: ' + str([pole]*multiplicator) + '\n'

    if (controller == 'PIFeedbackController' or controller == 'LSSController'):
        tmp += '   r0: 3\n'
    tmp += '\n'
    return tmp

def writeSensor(senName, paramVal):
    return '  sensor:' + '\n'\
            + '   type: ' + senName + '\n'\
            + '   delay: ' + str(paramVal) + '\n'\
            + '   output: ' + '[0,1,2,3]' + '\n\n'

def writeDisturbance(disName, paramVal):
    if paramVal == 0:
        return '  disturbance:' + '\n'\
            + '   type: None\n'

    return '  disturbance:' + '\n'\
            + '   type: ' + disName + '\n'\
            + '   mean value: ' + '0' + '\n'\
            + '   sigma: ' + str(paramVal) + '\n\n'

def saveOutput(output, cName, poles, parameter, limits):
    '''
        wrapper to save regime file
    '''
    dirPath = os.path.join(os.path.pardir, os.path.pardir, 'regimes', 'generated')

    if not os.path.isdir(dirPath):
        os.makedirs(dirPath)
    
    if cName == 'all':
        fileName = 'C1_' + cName\
                            + '_' + str(parameter)\
                            + '.sreg'
    else:
        fileName = 'C1_' + cName + '_poles' + str(poles) \
                        + '_' + str(parameter)\
                        + '(' + str(simLimits[0]) + ',' + str(simLimits[-1])\
                        + ').sreg'

    filePath = os.path.join(dirPath, fileName)
    with open(filePath, 'w') as f:
        f.write(output)

#-------------------------------------------------------------
# preamble
#-------------------------------------------------------------
preamble = '- name: C1-simulation-setup\n'\
+'  clear previous: !!python/bool True\n'\
+'\n'\
+'  model:\n'\
+'   type: BallBeamModel\n'\
+'\n'\
+'  solver:\n'\
+'   type: VODESolver\n'\
+'   end time: '+str(end_time)+'\n'\
+'\n'\
+'  trajectory:\n'\
+'   type: SmoothTransitionTrajectory\n'\
+'   Positions: [0, 3]\n'\
+'   start time: 0\n'\
+'   delta t: 5\n'\
+'\n'

#-------------------------------------------------------------
# main
#-------------------------------------------------------------
collection = ''
for controller in simCases:
    lines = preamble


    # how many poles?
    if controller == 'PIFeedbackController':
        multiplicator = 5
    else:
        multiplicator = 4

    #set correct poles
    poles = st.poles[controller]
    
    #get simulation limits
    limits = st.paramStabilityLimitsSigmaDelay[controller][parameter]
    #get simulation step size
    step_size = st.paramStabilityLimitsSigmaDelay[controller][parameter + '_step']
    
    #create simulationArray
    simLimits = np.arange(limits[0], limits[1] + step_size, step_size)
    
    #search limits
    for val in simLimits:
        lines += writeRegime(controller, poles, parameter, val, '')
        lines += writeController(controller, poles, multiplicator)
        if parameter == 'sigma':
            lines += writeDisturbance('GaussianNoiseDisturbance', val)
        if parameter == 'delay':
            lines += writeSensor('DeadTimeSensor', val)   
        lines += '\n\n'

        saveOutput(lines, controller, poles, parameter, simLimits)

    collection += lines

if len(simCases) > 1:
    saveOutput(collection, 'all', '', parameter, simLimits)

print('done.')



# -*- coding: utf-8 -*-
from collections import OrderedDict
import numpy as np

import pymoskito as pm


# class
class PendulumModel(pm.Model):
    public_settings = OrderedDict([("initial state", [0, 180.0, 0, 0]),
                                   ("cart mass", 4.3),  # [kg]
                                   ("cart friction", 10),  # [Ns/m]
                                   ("pendulum mass", 0.32),  # [kg]
                                   ("pendulum inertia", 0.07),  # [kg*m^2]
                                   ("pendulum friction", 0.03),  # [Nms]
                                   ("pendulum length", 0.35),  # [m]
                                   ("gravity", 9.81)])  # [m/s^2]

    # init
    def __init__(self, settings):
        # conversion from degree to radiant
        settings["initial state"][1] = np.deg2rad(settings["initial state"][1])
        settings["initial state"][3] = np.deg2rad(settings["initial state"][3])

        # add specific "private" settings
        settings.update(state_count=4)
        settings.update(input_count=1)
        settings.update({"output_info": {0: {"Name": "cart position",
                                             "Unit": "m"}}})
        pm.Model.__init__(self, settings)

    # state
    def state_function(self, t, x, args):
        # definitional
        s = x[0]
        phi = x[1]
        ds = x[2]
        dphi = x[3]
        F = args[0].squeeze()

        # shortcuts for readability
        M = self._settings["cart mass"]
        D = self._settings["cart friction"]
        m = self._settings["pendulum mass"]
        J = self._settings["pendulum inertia"]
        d = self._settings["pendulum friction"]
        l = self._settings["pendulum length"]
        g = self._settings["gravity"]

        dx1 = ds
        dx2 = dphi
        dx3 = ((J * F
                - J * D * ds
                - m * l * J * dphi ** 2 * np.sin(phi)
                + (m * l) ** 2 * g * np.sin(phi) * np.cos(phi)
                - m * l * d * dphi * np.cos(phi))
               / ((M + m) * J - (m * l * np.cos(phi)) ** 2))
        dx4 = ((m * l * np.cos(phi) * F
                - m * l * D * ds * np.cos(phi)
                - (m * l * dphi) ** 2 * np.cos(phi) * np.sin(phi)
                + (M + m) * m * l * g * np.sin(phi) - (M + m) * d * dphi)
               / ((M + m) * J - (m * l * np.cos(phi)) ** 2))

        dx = np.array([dx1, dx2, dx3, dx4])
        return dx

    # output
    def calc_output(self, input_vector):
        return input_vector[0]



# -*- coding: utf-8 -*-
import numpy as np
from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg as FigureCanvas
from matplotlib.figure import Figure
from matplotlib.lines import Line2D as Line

from .processing_core import PostProcessingModule, MetaProcessingModule
from .resources import colors
from .tools import sort_tree

__all__ = ["StepResponse", "PlotAll", "XYMetaProcessor"]


class StepResponse(PostProcessingModule):
    """
    Postprocessor that creates diagrams for step response experiments.

    Various measures are taken, displayed in a diagram and saved to result
    files. The following list contains the measured/calculated entries and
    their explanations:

        - rise time (t_rise): time needed until the output reaches 90% of its
        desired value
        - correction time (t_corr): time needed until the output reaches the
        desired value for the first time
        - overshoot time (t_over): time of the greatest overshot
        - overshot: maximum error between is and desired while approaching the
        desired value
        - damping time (t_damp): time needed for the output to enter and remain
        in an epsilon region around the desired value

    Args:
        model_idx(int): Index of the model output to use.
        trajectory_idx(int): Index of the trajectory output to use.

    """
    line_color = colors.HKS41K100
    line_width = 2
    line_style = '-'
    font_size = 20
    font_color = colors.HKS41K100
    spacing = 0.05
    counter = 0
    eps = 1e-3

    def __init__(self, model_idx=0, trajectory_idx=0):
        PostProcessingModule.__init__(self)
        self.model_idx = model_idx
        self.trajectory_idx = trajectory_idx
        self.label_positions = None
        return

    def run(self, data):

        # dict for calculated values
        output = {}

        # reset counter
        self.counter = 0

        # calculate data sets
        t = data["results"]["time"]
        y = data["results"]["Model"][:, self.model_idx]
        traj_data = data["results"]["Trajectory"]
        if traj_data.ndim == 2:
            yd = data["results"]["Trajectory"][-1, self.trajectory_idx]
        elif traj_data.ndim == 3:
            yd = data["results"]["Trajectory"][-1, self.trajectory_idx, 0]
        else:
            raise ValueError("unknown Trajectory type.")

        self.label_positions = np.arange(y.min() + 0.1 * yd,
                                         yd,
                                         (yd - y.min()) / 4)

        # create plot
        fig = Figure()
        axes = fig.add_subplot(111)
        axes.grid()
        axes.set_title(r"Step Response")
        axes.plot(t, y, c=colors.HKS44K100, linewidth=self.line_width)
        axes.set_xlim(left=0, right=t[-1])
        axes.set_xlabel(r"Time in $\si{\second}$")
        axes.set_ylabel(r"System Output in $\si{\metre}$")

        # create desired line
        desired_line = Line([0, t[-1]], [yd, yd],
                            lw=1, ls=self.line_style, c='k')
        axes.add_line(desired_line)

        # calc rise-time (Anstiegszeit)
        try:
            t_rise = t[np.where(y > 0.9 * yd)][0]
            self.create_time_line(axes, t, y, t_rise, r"$T_r$")
            output.update({"t_rise": t_rise})
        except IndexError:
            output.update({"t_rise": None})

        # calc correction-time (Anregelzeit)
        try:
            t_corr = t[np.where(y > yd)][0]
            self.create_time_line(axes, t, y, t_corr, r"$T_{c}$")
            output.update({"t_corr": t_corr})
        except IndexError:
            output.update({"t_corr": None})

        # calc overshoot-time and overshoot in percent
        # (Überschwingzeit und Überschwingen)
        if output["t_corr"]:
            if yd > 0:
                y_max = np.max(y[np.where(t > output["t_corr"])])
            else:
                y_max = np.min(y[np.where(t > output["t_corr"])])

            t_over = t[np.where(y == y_max)][0]
            overshoot = y_max - yd
            overshoot_per = overshoot / yd * 100
            self._logger.info("Overshoot: {}%".format(overshoot_per))

            self.create_time_line(axes, t, y, t_over, r"$T_o$")
            output.update(dict(t_over=t_over,
                               overshoot=overshoot,
                               overshoot_percent=overshoot_per))
        else:
            output.update(dict(t_over=None,
                               overshoot=None,
                               overshoot_percent=None))

        # calc damping-time (Beruhigungszeit)
        try:
            enter_idx = -1
            for idx, val in enumerate(y):
                if enter_idx == -1:
                    if abs(val - yd) < self.eps:
                        enter_idx = idx
                else:
                    if abs(val - yd) >= self.eps:
                        enter_idx = -1

            t_damp = t[enter_idx]
            self.create_time_line(axes, t, y, t_damp, r"$T_{\epsilon}$")
            output.update({"t_damp": t_damp})
        except IndexError:
            output.update({"t_damp": None})

        # create epsilon tube
        upper_bound_line = Line([0, t[-1]], [yd + self.eps, yd + self.eps],
                                ls="--", c=self.line_color)
        axes.add_line(upper_bound_line)
        lower_bound_line = Line([0, t[-1]], [yd - self.eps, yd - self.eps],
                                ls="--", c=self.line_color)
        axes.add_line(lower_bound_line)

        # calc stationary control deviation
        control_deviation = y[-1] - yd
        output.update({"stationary_error": control_deviation})
        # check for sim success
        if not data["results"]["finished"]:
            for key in output.keys():
                output[key] = None

        # add settings and metrics to dictionary results
        results = {}
        y_des = np.ones_like(y) * yd
        step_width = 1 / data["modules"]["Solver"]["measure rate"]
        results.update({"metrics": self.calc_metrics(y,
                                                     y_des,
                                                     step_width,
                                                     output)})
        results.update({"modules": data["modules"]})

        canvas = FigureCanvas(fig)

        self.write_output_files(data["regime name"], fig, results)

        return [dict(name='_'.join([data["regime name"], self.name]),
                     figure=canvas)]

    def create_time_line(self, axes, t, y, time_value, label):
        if time_value != t[-1]:
            time_line = Line([time_value, time_value],
                             [np.min(y), y[np.where(t == time_value)][0]],
                             linewidth=self.line_width,
                             ls=self.line_style,
                             c=self.line_color)
            axes.add_line(time_line)
            axes.text(time_value + self.spacing,
                      self.label_positions[self.counter],
                      label,
                      color=self.font_color,
                      size=self.font_size)
            self.counter += 1

    def calc_metrics(self, measured_values, desired_values, step_width, output):
        """
        Calculate metrics

        Args:
            measured_values(array): Array holding the measured outputs.
            desired_values(array): Array holding the desired outputs.
            step_width(float): Simulation step width.
            output(dict): Values computed by this module.

        Returns:
            dict: Calculated metrics.
        """
        l1_norm_itae = self.calc_l1_norm_itae(measured_values,
                                              desired_values,
                                              step_width)
        l1_norm_abs = self.calc_l1_norm_abs(measured_values,
                                            desired_values,
                                            step_width)

        self._logger.info("L1NormITAE: {}".format(l1_norm_itae))
        self._logger.info("L1NormAbs: {}".format(l1_norm_abs))

        return {'L1NormITAE': l1_norm_itae, 'L1NormAbs': l1_norm_abs}


class PlotAll(PostProcessingModule):
    """
    plot diagrams of all system quantities
    """

    def __init__(self):
        PostProcessingModule.__init__(self)
        return

    def run(self, data):

        # dict for calculated values
        output = {}
        return_list = []

        t = data["results"]["time"]
        val = t  # default

        for module_name, module_data in data["results"].items():
            if module_name in ["time", "finished", "Simulation"]:
                continue

            module_shape = module_data.shape
            for idx in range(module_shape[1]):
                if len(module_shape) == 3:
                    val = module_data[:, idx, 0]
                if len(module_shape) == 2:
                    val = module_data[:, idx]

                plot_name = "_".join([data["regime name"],
                                      self.name,
                                      module_name, str(idx)])
                fig = Figure()
                axes = fig.add_subplot(111)
                axes.set_title(r"\textbf{"
                               + "{} {}".format(
                                    module_name.replace("_", " "), idx)
                               + "}")
                axes.plot(t, val, c='k')
                axes.set_xlim(left=0, right=t[-1])
                axes.set_xlabel(r"Time in s")
                axes.set_ylabel(r"{} {}".format(
                    module_name.replace("_", " "), idx))
                axes.grid(True)
                canvas = FigureCanvas(fig)

                return_list.append({"name": plot_name, "figure": canvas})

        return return_list


class XYMetaProcessor(MetaProcessingModule):
    """
    Create XY-diagrams for the given keys from metadata.

    Args:
        x_path(list): List of dictionary keys pointing to the desired values
            of the x axis.
        y_path(list): List of dictionary keys pointing to the desired values
            of the y axis.
        sort_key(list): List of dictionary keys used for grouping the meta data.
        x_idx(int): If the *x_path* points to an array but only one component
            is to be selected, its index can be given. [-1]
        y_idx(int): If the *y_path* points to an array but only one component
            is to be selected, its index can be given. [-1]
        line_style(str): Either ``line`` or ``bar``.
    """

    def __init__(self, x_path, y_path, sort_key, x_idx=-1, y_idx=-1,
                 line_style="line", title="", x_label="", y_label=""):
        MetaProcessingModule.__init__(self)
        self.x_path = x_path
        self.x_idx = x_idx
        self.y_path = y_path
        self.y_idx = y_idx
        self.sort_key = sort_key
        self.line_style = line_style
        self.title = title
        self.x_lbl = x_label
        self.y_lbl = y_label

        self.fig = None
        self.axes = None

        return

    def process(self, post_results):
        # create tree with relevant data
        source = sort_tree(post_results, self.sort_key)

        # create plot
        self.fig = Figure()
        self.axes = self.fig.add_subplot(111)

        self.plot_family(source,
                         self.x_path,
                         self.y_path,
                         self.line_style,
                         self.x_idx,
                         self.y_idx)
        self.set_plot_labeling(self.title, True, self.x_lbl, self.y_lbl,
                               self.line_style)

        # extract member_names, therefore
        # (subtract common appendix like *Controller or *Feedforward)
        member_names = [x[:-len(self.x_path[1])] for x in source.keys()]

        canvas = FigureCanvas(self.fig)

        # write output files
        file_name = (self.name
                     + "_".join([self.x_path[1], "("])
                     + "".join(member_names) + ")")
        self.write_output_files(file_name, self.fig)

        return [{'figure': canvas, 'name': self.name}]


def construct_result_dict(data, output):
    # check for sim success
    if not data["results"]["finished"]:
        for key in output.keys():
            output[key] = None

    # add settings and metrics to dictionary results
    results = {}
    results.update({'metrics': output})
    results.update({'modules': data['modules']})

    return results



# -*- coding: utf-8 -*-
import logging.config
import os
import yaml

# make all plotting libs use qt5
import matplotlib as mpl
mpl.use('Qt5Agg')
os.environ["PYQTGRAPH_QT_LIB"] = "PyQt5"

# enable high dpi scaling
os.environ["QT_AUTO_SCREEN_SCALE_FACTOR"] = "1"

from .registry import *

from .processing_core import *
from .processing_gui import *
from .generic_processing_modules import *

from .simulation_gui import *
from .simulation_modules import *
from .generic_simulation_modules import *

from .binding_modules import *

from .tools import *
from .controltools import *
from .visualization import *
from .resources import *

__author__ = 'Stefan Ecklebe'
__email__ = 'stefan.ecklebe@tu-dresden.de'
__version__ = '0.4.0'

# configure logging
with open(get_resource("logging.yaml", ""), "r") as f:
    log_conf = yaml.full_load(f)

logging.config.dictConfig(log_conf)

# register all generic modules
register_simulation_module(Model, LinearStateSpaceModel)
register_simulation_module(Solver, ODEInt)
register_simulation_module(Trajectory, SmoothTransition)
register_simulation_module(Trajectory, HarmonicTrajectory)
register_simulation_module(Trajectory, Setpoint)
register_simulation_module(Controller, PIDController)
register_simulation_module(Controller, LinearStateSpaceController)
register_simulation_module(ModelMixer, AdditiveMixer)
register_simulation_module(ObserverMixer, AdditiveMixer)
register_simulation_module(Limiter, ModelInputLimiter)
register_simulation_module(Sensor, DeadTimeSensor)
register_simulation_module(Disturbance, GaussianNoise)

register_processing_module(PostProcessingModule, PlotAll)



# -*- coding: utf-8 -*-
from collections import OrderedDict
import numpy as np

import pymoskito as pm

from . import settings as st


class CarModel(pm.Model):
    public_settings = OrderedDict([('initial state', st.initial_state),
                                   ("d1", st.d1),
                                   ("l2", st.l2),
                                   ("d2", st.d2),
                                   ("l3", st.l3),
                                   ("dia", st.dia),
                                   ("car_radius", st.car_radius),
                                   ("wheel", st.wheel),
                                   ])

    def __init__(self, settings):
        # add specific "private" settings
        settings.update(state_count=5)
        settings.update(input_count=2)
        settings.update({"output_info": {
            0: {"Name": "x-position", "Unit": "m"},
            1: {"Name": "y-position", "Unit": "m"},
            2: {"Name": "car-angle", "Unit": "rad"},
            3: {"Name": "1st-trailer-angle", "Unit": "rad"},
            4: {"Name": "2nd-trailer-angle", "Unit": "rad"},
        }})
        pm.Model.__init__(self, settings)

        # shortcuts for readability
        self.d1 = self._settings['d1']
        self.l2 = self._settings['l2']
        self.d2 = self._settings['d2']
        self.l3 = self._settings['l3']

    def state_function(self, t, x, args):
        """
        Calculations of system state changes
        :param x: state
        :param t: time
        :type args: system input force on the cart
        """

        x1, x2, x3, x4, x5 = x
        v, w = 1, 1

        return np.array([v * np.cos(x3),
                         v * np.sin(x3),
                         w,
                         v / self.l2 * np.sin(x3 - x4) - self.d1 / self.l2 * w * np.cos(x3 - x4),
                         v * (1 / self.l3 * np.sin(x3 - x5) - (self.d2 + self.l2) / (self.l2 * self.l3) * np.sin(
                             x3 - x4) * np.cos(x4 - x5)) \
                         + w * (-self.d1 / self.l3 * np.cos(x3 - x5) + (self.d1 * (self.l2 + self.d2)) / (
                             self.l2 * self.l3) * np.cos(x3 - x4) * np.cos(x4 - x5))
                         ])

    def root_function(self, x):
        return [False]

    def check_consistency(self, x):
        pass

    def calc_output(self, input):
        return input


pm.register_simulation_module(pm.Model, CarModel)



#!/usr/bin/python2
# -*- coding: utf-8 -*-

import numpy as np
import os
import settings as st

controllerList = ['FController', 'GController', 'JController',\
                'LSSController', 'PIFeedbackController']
               
#TODO: überlege, in welchem Punkt Regler linearisiert werden und dazu Pole
#####################################################
# Einstellungen               
number = 2

fRange = [0.05, 0.2]
fStepSize = 0.01
ampl = 3

#####################################################
lines = ''

# load head file
filePath = os.path.join('A2_head.sray')
with open(filePath, 'r') as f:
    head = f.read()

for contr in controllerList:
    # Pole festlegen nach A1
    pol = st.poles[contr]
    print(contr)
    print('pol:', pol)
    
    if contr == 'PIFeedbackController':
        poles = [pol, pol, pol, pol, pol]
    else:
        poles = [pol, pol, pol, pol]
    
    
    
    fs = np.arange(fRange[0],fRange[1] + fStepSize, fStepSize)
    lines += '\n'
    
    for f in fs:
            lines += '- name: A2_' + contr + '_A' + str(ampl) + '_f' + str(f) + '\n'
            lines += '  clear previous: !!python/bool False' + '\n'
            lines += '\n'
            lines += '  solver: ' + '\n'
            lines += '   type: VODESolver' + '\n'
            lines += '   initial state: ' + str([ampl, 0, 0, 0]) + '\n'
            lines += '\n'
            lines += '  controller: ' + '\n'
            lines += '   type: '  + contr + '\n'
            lines += '   poles: ' + str(poles) + '\n'
            lines += '\n'
            lines += '  trajectory:' +'\n'
            lines += '   type: HarmonicTrajectory' + '\n'
            lines += '   Amplitude: ' + str(ampl) + '\n'
            lines += '   Frequency: ' + str(f) + '\n'
            lines += '\n'
               
#write results
#print os.path.pardir
dirPath = os.path.join(os.path.pardir, os.path.pardir, 'regimes', 'generated')
if not os.path.isdir(dirPath):
    os.makedirs(dirPath)

fileName = 'A2_' + '_A('+str(ampl)+')_r0sweep' + '_all_fRange(' + str(fRange[0])\
            + ',' + str(fRange[1]) + ')'+'.sreg'
            
filePath = os.path.join(dirPath, fileName)   

with open(filePath, 'w') as f:
    f.write(head)    
    f.write(lines)
    
    
    
    



# -*- coding: utf-8 -*-
"""
Linearisation of the Ball and Beam system
"""

import sympy as sp
import numpy as np

from . import settings as st


def linearise_system(x0: np.ndarray, tau0: float) -> tuple:
    r"""
    Linearise the system at the steady state :math:`(\boldsymbol{x}^e, \tau^e)`
    
    Args:
        x0: Steady state.
        tau0: Corresponding system input.

    Returns:
        Tuple of matrices: Linear state space representation of the system.
    """
    params = sp.symbols('x, x1, x2, x3, x4, u1, tau, M , G , J , J_ball , R, B')
    x, x1, x2, x3, x4, u1, tau, M, G, J, J_ball, R, b = params

    x01 = x0[0]
    x02 = x0[1]
    x03 = x0[2]
    x04 = x0[3]

    # f1 = ddr
    f1 = b * (x1 * x4 ** 2 - G * sp.sin(x3))

    # f2 = ddtheta
    f2 = (tau - 2 * M * x1 * x2 * x4 - M * G * x1 * sp.cos(x3)) / (M * x1 ** 2
                                                                   + J + J_ball)

    f1 = sp.Matrix([f1])
    f2 = sp.Matrix([f2])

    x = sp.Matrix([[x1], [x2], [x3], [x4]])
    u = sp.Matrix([[u1, tau]])

    jac_a1 = f1.jacobian(x)
    jac_a2 = f2.jacobian(x)
    jac_b1 = f1.jacobian(u)
    jac_b2 = f2.jacobian(u)

    subs_list = [(x1, x01), (x2, x02), (x3, x03), (x4, x04), (tau, tau0),
                 (b, st.B), (J, st.J), (J_ball, st.Jb), (M, st.M), (G, st.G),
                 (R, st.R)]
    jac_a1 = jac_a1.subs(subs_list)
    jac_a2 = jac_a2.subs(subs_list)
    jac_b1 = jac_b1.subs(subs_list)
    jac_b2 = jac_b2.subs(subs_list)

    # system matrix A
    a = sp.Matrix([[0, 1, 0, 0],
                   [jac_a1[0], jac_a1[1], jac_a1[2], jac_a1[3]],
                   [0, 0, 0, 1],
                   [jac_a2[0], jac_a2[1], jac_a2[2], jac_a2[3]]
                   ])

    # input matrix B
    b = sp.Matrix([[jac_b1[0]],
                   [jac_b1[1]],
                   [jac_b2[0]],
                   [jac_b2[1]]
                   ])
    # output matrix C
    c = sp.Matrix([[1, 0, 0, 0]])

    # and we are done with sympy
    a_mat = np.array(a).astype(float)
    b_mat = np.array(b).astype(float)
    c_mat = np.array(c).astype(float)

    return a_mat, b_mat, c_mat





import numpy as np
from collections import OrderedDict

import pymoskito as pm

from . import settings as st


class BallInTubeModel(pm.Model):
    """
    Model of the Ball in Tube System
    """
    public_settings = OrderedDict([('initial state', st.initial_state),
                                   ('d_B', st.d_B),
                                   ('d_R', st.d_R),
                                   ('tube_length', st.tube_length),
                                   ('k_L', st.k_L),
                                   ('k_V', st.k_V),
                                   ('k_s', st.k_s),
                                   ('d', st.d),
                                   ('T', st.T),
                                   ('m', st.m),
                                   ('g', st.g),
                                   ])

    def __init__(self, settings):
        # add specific "private" settings
        settings.update(state_count=4)
        settings.update(input_count=1)
        settings.update({"output_info": {
            0: {"Name": "ball elevation", "Unit": "m"},
        }})
        pm.Model.__init__(self, settings)

        # shortcuts for readability
        self.d_B = self._settings['d_B']
        self.d_R = self._settings['d_R']
        self.tube_length = self._settings['tube_length']
        self.k_L = self._settings['k_L']
        self.k_V = self._settings['k_V']
        self.k_s = self._settings['k_s']
        self.d = self._settings["d"]
        self.T = self._settings['T']
        self.m = self._settings['m']
        self.g = self._settings['g']
        self.A_B = np.pi*self.d_B**2/4
        self.A_R = np.pi*self.d_R**2/4
        self.A_Sp = self.A_R - self.A_B

    def state_function(self, t, x, args):
        """
        Calculations of system state changes
        :param x: state
        :param t: time
        :type args: system input u
        """

        # definitional
        x1 = x[0]
        x2 = x[1]
        # x3 = x[2] is not used
        x4 = x[3]
        u = args[0].squeeze()

        dx1 = x2
        dx2 = -x1/self.T**2 - 2*self.d*x2/self.T + self.k_s*u*st.Vcc/(255*self.T**2)
        dx3 = x4
        dx4 = (self.k_L*((self.k_V*x1 - self.A_B*x4)/self.A_Sp)**2 - self.m*self.g)/self.m

        return np.array([dx1, dx2, dx3, dx4])

    def root_function(self, x):
        """
        in this case this means zero crossing detection for the balls elevation.
        :param x: state
        """
        x0 = x
        flag = False

        if x[2] <= 0:
            x0[2] = 0
            x0[3] = 0
            flag = True

        if x[0] <= 0:
            x0[0] = 0
            x0[1] = 0
            flag = True

        return flag, x0

    def check_consistency(self, x):
        """
        Checks if the model rules are violated
        :param x: state
        """
        if x[2] > (self._settings['tube_length']):  # + self._settings['tube_length']*0.2):
            raise pm.ModelException('Ball flew out of the tube')

    def calc_output(self, input_vector):
        """
        return ball position as output
        :param input_vector: input values
        :return: ball position
        """
        return input_vector[2]


class BallInTubeSpringModel(pm.Model):
    """
    Implementation of the Ball in Tube System with a spring on the bottom
    """
    public_settings = OrderedDict([('initial state', st.initial_state),
                                   ('d_B', st.d_B),
                                   ('d_R', st.d_R),
                                   ('tube_length', st.tube_length),
                                   ('k_L', st.k_L),
                                   ('k_V', st.k_V),
                                   ('k_s', st.k_s),
                                   ('d', st.d),
                                   ('T', st.T),
                                   ('m', st.m),
                                   ('g', st.g),
                                   ('K', st.K),
                                   ('D', st.D),
                                   ])

    def __init__(self, settings):
        # add specific "private" settings
        settings.update(state_count=4)
        settings.update(input_count=1)
        pm.Model.__init__(self, settings)

        # shortcuts for readability
        self.d_B = self._settings['d_B']
        self.d_R = self._settings['d_R']
        self.tube_length = self._settings['tube_length']
        self.k_L = self._settings['k_L']
        self.k_V = self._settings['k_V']
        self.k_s = self._settings['k_s']
        self.d = self._settings['d']
        self.T = self._settings['T']
        self.m = self._settings['m']
        self.g = self._settings['g']
        self.K = self._settings['K']
        self.D = self._settings['D']
        self.A_B = np.pi*self.d_B**2/4
        self.A_R = np.pi*self.d_R**2/4
        self.A_Sp = self.A_R - self.A_B

    def state_function(self, t, x, args):
        """
        Calculations of system state changes
        :param x: state
        :param t: time
        :type args: system input u
        """

        # definitional
        x1 = x[0]
        x2 = x[1]
        x3 = x[2]
        x4 = x[3]
        u = args[0].squeeze()

        dx1 = x2
        dx2 = -x1/self.T**2 - 2*self.d*x2/self.T + self.k_s*u*12/(255*self.T**2)
        dx3 = x4

        if x3 < 0:
            # realize a extra spring with spring stiffness K and spring damping D
            dx4 = -(self.K*x3)/self.m - (self.D*x4)/self.m \
                  + (self.k_L*((self.k_V*x1 - self.A_B*x4)/self.A_Sp)**2)/self.m - self.g
        else:
            dx4 = (self.k_L*((self.k_V*x1 - self.A_B*x4)/self.A_Sp)**2)/self.m - self.g

        return np.array([dx1, dx2, dx3, dx4])

    def root_function(self, x):
        """
        in this case this means zero crossing detection for the balls elevation.
        :param x: state
        """
        x0 = x
        flag = False

        # fan speed
        if x[0] <= 0:
            x0[0] = 0
            x0[1] = 0
            flag = True

        return flag, x0

    def check_consistency(self, x):
        """
        Checks if the model rules are violated
        :param x: state
        """
        if x[2] > (self._settings['tube_length']):  # + self._settings['tube_length']*0.2):
            raise pm.ModelException('Ball flew out of the tube')

    def calc_output(self, input_vector):
        """
        return ball position as output
        :param input_vector: input values
        :return: ball position
        """
        return input_vector[2]


pm.register_simulation_module(pm.Model, BallInTubeModel)
pm.register_simulation_module(pm.Model, BallInTubeSpringModel)



"""
test_simulation_gui
----------------------------------

Tests for `simulation_gui` module.
"""

import sys
import unittest

from PyQt5.QtCore import Qt
from PyQt5.QtTest import QSignalSpy, QTest
from PyQt5.QtWidgets import QApplication

from pymoskito import SimulationGui

app = QApplication([])


class TestGUIBasics(unittest.TestCase):

    def setUp(self):
        self.gui = SimulationGui()
        # self.gui.show()

    def test_defaults(self):
        # actions - enable at start
        self.assertTrue(self.gui.actLoadRegimes.isEnabled())
        # self.assertTrue(self.gui.actLoadStandardState.isEnabled())
        self.assertTrue(self.gui.actSimulateCurrent.isEnabled())
        self.assertTrue(self.gui.actSlow.isEnabled())
        self.assertTrue(self.gui.actFast.isEnabled())

        # actions - disabled at start
        self.assertFalse(self.gui.actSave.isEnabled())
        self.assertFalse(self.gui.actSimulateAll.isEnabled())
        self.assertFalse(self.gui.actPlayPause.isEnabled())
        self.assertFalse(self.gui.actStop.isEnabled())
        try:
            import vtk
            self.assertTrue(self.gui.actResetCamera.isEnabled())
        except ImportError:
            self.assertFalse(self.gui.actResetCamera.isEnabled())

        # timeline
        self.assertFalse(self.gui.timeSlider.isEnabled())

        # playback speed
        self.assertTrue(self.gui.speedControl.isEnabled())
        self.assertEqual(self.gui.speedControl.value(), 6)

    def tearDown(self):
        self.gui.close()


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
import numpy as np
from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg as FigureCanvas
from matplotlib.figure import Figure

import pymoskito as pm


class TwoPendulum(pm.PostProcessingModule):
    """
    plot diagrams of all system quantities
    """
    def __init__(self):
        pm.PostProcessingModule.__init__(self)
        return

    def run(self, data):
        return_list = []

        t = data["results"]["time"]
        x0 = data["results"]["Solver"][:, 0]
        x0_vel = data["results"]["Solver"][:, 1]
        phi1 = data["results"]["Solver"][:, 2]
        phi1_vel = data["results"]["Solver"][:, 3]
        phi2 = data["results"]["Solver"][:, 4]
        phi2_vel = data["results"]["Solver"][:, 5]

        val_list = [x0, x0_vel, phi1, phi1_vel, phi2, phi2_vel]

        title_list = ["Position of the cart",
                      "Velocity of the cart",
                      "Angle long pendulum",
                      "Angular velocity long pendulum",
                      "Angle short pendulum",
                      "Angular velocity short pendulum"]
        x_label_list = [r"$x_{0}$ in m",
                        r"$\dot{x}_{0}$ in m/s",
                        r"$\varphi_{1}$ in degree",
                        r"$\dot{\varphi}_{1}$ in degree/s",
                        r"$\varphi_{2}$ in degree",
                        r"$\dot{\varphi}_{2}$ in degree/s"]
        filename_list = ["x0",
                         "x0_vel",
                         "phi1",
                         "phi1_vel",
                         "phi2",
                         "phi2_vel"]
        if 0:
            for idx, val in enumerate(val_list):
                fig = Figure()
                axes = fig.add_subplot(111)
                axes.set_title(r"\textbf{%s}" % title_list[idx])
                axes.plot(t, val, c="k")
                axes.set_xlim(left=0, right=t[-1])
                axes.set_xlabel(r"Time in s")
                axes.set_ylabel(r"%s" % x_label_list[idx])
                axes.grid(True)
                canvas = FigureCanvas(fig)

                plot_name = "_".join([data["regime name"], self.name, filename_list[idx]])
                return_list.append({"name": plot_name, "figure": canvas})

                # save file
                self.write_output_files(result_name="_".join([data["regime name"], filename_list[idx]]),
                                        figure=fig,
                                        output=pm.construct_result_dict(data, output={}))

        #################################################
        # this section is for combined and custom plots
        # plot both angle of the pendulums in one plot
        plot_selection = {"x0_x0_vel": True,
                          "phi1_phi2": True,
                          "V_V_dot": True}

        if plot_selection["x0_x0_vel"]:
            self._logger.info("creating x0_x0_vel plot")

            fig = Figure()
            axes = fig.add_subplot(111)
            axes.plot(t, x0, c="b", label=r"$x_{0}$ in m")
            axes.plot(t, x0_vel, c="r", label=r"$\dot{x}_{0}$ in $\frac{\mathrm{m}}{\mathrm{s}}$")
            axes.set_xlim(left=t[0], right=t[-1])
            axes.set_xlabel(r"Time in s")
            axes.set_ylabel(r"")

            axes.grid(True)
            axes.legend()
            axes.legend(loc=0)
            canvas = FigureCanvas(fig)

            plot_name = "_".join([data["regime name"], self.name, "x0_x0_vel"])
            return_list.append({"name": plot_name, "figure": canvas})

            # save file
            self.write_output_files(result_name="_".join([data["regime name"], "x0_x0_vel"]),
                                    figure=fig,
                                    output=pm.construct_result_dict(data, output={}))

        if plot_selection["phi1_phi2"]:
            self._logger.info("creating phi1_phi2 plot")

            fig = Figure()
            axes = fig.add_subplot(111)
            axes.plot(t, np.rad2deg(phi1), c="k", label=r"$\varphi_{1}$")
            axes.plot(t, np.rad2deg(phi2), c="b", label=r"$\varphi_{2}$")
            axes.set_xlim(left=t[0], right=t[-1])
            axes.set_xlabel(r"Time in s")
            axes.set_ylabel(r"Angle in $^{\circ}$")

            # start, end = axes.get_ylim()
            # fine segmentation
            # axes.set_yticks(np.arange(int(start/30)*30, int(end/30)*30 + 30, 30), minor=True)
            # axes.set_yticks(np.arange(int(start/90)*90, int(end/90)*90 + 90, 90), minor=False)
            # rough segmentation
            # axes.set_yticks(np.arange(int(start/90)*90, int(end/90)*90 + 90, 90), minor=True)
            # axes.set_yticks(np.arange(int(start/180)*180, int(end/180)*180 + 180, 180), minor=False)
            # axes.grid(which="minor", alpha=0.2)
            # axes.grid(which="major", alpha=0.5)
            axes.grid(True)
            axes.legend()
            axes.legend(loc=0)
            canvas = FigureCanvas(fig)

            plot_name = "_".join([data["regime name"], self.name, "phi1_phi2"])
            return_list.append({"name": plot_name, "figure": canvas})

            # save file
            self.write_output_files(result_name="_".join([data["regime name"], "phi1_phi2"]),
                                    figure=fig,
                                    output=pm.construct_result_dict(data, output={}))

        if plot_selection["V_V_dot"]:
            self._logger.debug("creating V_V_dot plot")

            V, V_dot = self.calc_v_and_v_dot(data, val_list)

            fig = Figure()
            axes = fig.add_subplot(111)
            axes.plot(t, V, c="k", label=r"$V$ in $\frac{\mathrm{kg}^{2}\mathrm{m}^{4}}{\mathrm{s}^{4}}$")
            # axes.plot(t, V_dot, c="b", label=r"$\dot{V}$ in $\frac{\mathrm{kg}^{2}\mathrm{m}^{4}}{\mathrm{s}^{5}}$")
            axes.set_xlim(left=t[0], right=t[-1])
            axes.set_xlabel(r"Time in s")
            axes.set_ylabel(r"")

            axes.grid(True)
            axes.legend()
            axes.legend(loc=0)
            canvas = FigureCanvas(fig)

            plot_name = "_".join([data["regime name"], self.name, "V_V_dot"])
            return_list.append({"name": plot_name, "figure": canvas})

            # save file
            self.write_output_files(result_name="_".join([data["regime name"], "V_V_dot"]),
                                    figure=fig,
                                    output=pm.construct_result_dict(data, output={}))

        return return_list

    @staticmethod
    def calc_v_and_v_dot(data, val_list):
        if "Controller" in data["modules"]:
            if "long pendulum" in data["modules"]["Controller"] and "short pendulum" in data["modules"]["Controller"]:
                # extract parameter to simplify the calculation of V and V_dot
                m0_star = data["modules"]["Model"]["m0*"]
                m1_star = data["modules"]["Model"]["m1*"]
                m2_star = data["modules"]["Model"]["m2*"]
                l1_star = data["modules"]["Model"]["l1*"]
                l2_star = data["modules"]["Model"]["l2*"]
                # d0 = data["modules"]["Model"]["d0"]
                d1 = data["modules"]["Model"]["d1"]
                d2 = data["modules"]["Model"]["d2"]
                J_DP1 = data["modules"]["Model"]["J_DP1"]
                J_DP2 = data["modules"]["Model"]["J_DP2"]
                g = data["modules"]["Model"]["g"]
                k = data["modules"]["Controller"]["k"]

                # calculate point mass model parameter
                l1 = J_DP1/(m1_star*l1_star)
                l2 = J_DP2/(m2_star*l2_star)

                m1 = (m1_star*l1_star)**2/J_DP1
                m2 = (m2_star*l2_star)**2/J_DP2
                m0 = m0_star + (m1_star - m1) + (m2_star - m2)

                w = m1*l1/(m2*l2)

                # extract state from val_list
                x0, x0_vel, phi1, phi1_vel, phi2, phi2_vel = val_list
                E0 = 0.5*m0*x0**2
                E1 = 0
                E2 = 0

                if data["modules"]["Controller"]["long pendulum"] == "u":
                    E1 = 0.5*m1*l1**2*phi1_vel**2 + m1*g*l1*(np.cos(phi1) + 1)
                elif data["modules"]["Controller"]["long pendulum"] == "o":
                    E1 = 0.5*m1*l1**2*phi1_vel**2 + m1*g*l1*(np.cos(phi1) - 1)

                if data["modules"]["Controller"]["short pendulum"] == "u":
                    E2 = 0.5*m2*l2**2*phi2_vel**2 + m2*g*l2*(np.cos(phi2) + 1)
                elif data["modules"]["Controller"]["short pendulum"] == "o":
                    E2 = 0.5*m2*l2**2*phi2_vel**2 + m2*g*l2*(np.cos(phi2) - 1)

                G = m0*x0_vel*E0 + m1*l1*phi1_vel*np.cos(phi1)*E1 + m2*l2*phi2_vel*np.cos(phi2)*E2*w**2

                V = 0.5*E0**2 + 0.5*E1**2 + 0.5*E2**2

                V_dot = -k*G**2 - d1*phi1_vel**2*E1 - d2*phi2_vel**2*E2

        else:
            V = np.zeros(len(val_list[0]))
            V_dot = np.zeros(len(val_list[0]))

        return [V, V_dot]


class SimulationVerification(pm.PostProcessingModule):
    """
    plot diagrams of all system quantities
    """

    def __init__(self):
        pm.PostProcessingModule.__init__(self)
        return

    def run(self, data):
        return_list = []

        t = data["results"]["time"]
        x0 = data["results"]["Solver"][:, 0]
        x0_vel = data["results"]["Solver"][:, 1]
        phi1 = data["results"]["Solver"][:, 2]
        phi1_vel = data["results"]["Solver"][:, 3]
        phi2 = data["results"]["Solver"][:, 4]
        phi2_vel = data["results"]["Solver"][:, 5]

        # extract parameter for further calculation
        m0_star = data["modules"]["Model"]["m0*"]
        m1_star = data["modules"]["Model"]["m1*"]
        m2_star = data["modules"]["Model"]["m2*"]
        l1_star = data["modules"]["Model"]["l1*"]
        l2_star = data["modules"]["Model"]["l2*"]
        d0 = data["modules"]["Model"]["d0"]
        d1 = data["modules"]["Model"]["d1"]
        d2 = data["modules"]["Model"]["d2"]
        J_DP1 = data["modules"]["Model"]["J_DP1"]
        J_DP2 = data["modules"]["Model"]["J_DP2"]
        g = data["modules"]["Model"]["g"]

        # calculate point mass model parameter
        l1 = J_DP1/(m1_star*l1_star)
        l2 = J_DP2/(m2_star*l2_star)
        m1 = (m1_star*l1_star)**2/J_DP1
        m2 = (m2_star*l2_star)**2/J_DP2
        m0 = m0_star + (m1_star - m1) + (m2_star - m2)

        # calculate the analytic solution of a pendulum angle
        # initial equation
        # phi_dd = g*sin(phi)/l1 - (d1*phi_d)/(m1*l1**2) + cos(phi)*u/l1
        # linearisation in point phi_e = pi and u = 0 --> transformation: phi_lin = phi - phi_e
        phi_e = np.pi
        phi_vel_e = 0
        # phi_dd_lin  + (d1*phi_d_lin)/(m1*l1**2) + g*phi_lin/l1
        # usage of the appropriation method
        # lambda**2 + (d1/(m1*l1**2))*lambda + g/l = 0
        # usage of the pq-formula
        p = d1/(m1*l1**2)
        q = g/l1
        # parameter of the conjugate complex zero pair
        a = -p/2
        b = np.sqrt(abs((p/2)**2 - q))

        # get initial values for angle and angle velocity
        phi1_0 = phi1[0]
        phi1_0_vel = phi1_vel[0]
        # transform initial values into linear valid range
        phi1_0_lin = phi1_0 - phi_e
        phi1_0_vel_lin = phi1_0_vel - phi_vel_e

        # calculate unknown parameter of the analytic solution
        C1 = phi1_0_lin
        C2 = -a*phi1_0_lin/b

        phi1_lin_analytic = np.exp(a*t)*(C1*np.cos(b*t) + C2*np.sin(b*t))
        phi1_vel_lin_analytic = np.exp(a*t)*(a*(C1*np.cos(b*t) + C2*np.sin(b*t)) + b*(-C1*np.sin(b*t) + C2*np.cos(b*t)))

        # re-transformation of phi with phi = phi_lin + phi_e
        phi1_analytic = phi1_lin_analytic + phi_e
        phi1_vel_analytic = phi1_vel_lin_analytic + phi_vel_e

        self._logger.debug("creating plot for verification")
        self._logger.debug("error of phi1 and phi1 analytic: {}".format(sum(abs(phi1-phi1_analytic))))
        self._logger.debug("Parameter of the analytic solution: a={}, b={}, C1={}, C2={}".format(a, b, C1, C2))

        fig = Figure()
        axes = fig.add_subplot(211)
        axes.plot(t, np.rad2deg(phi1),
                  c="k",
                  linestyle="-",
                  label=r"$\varphi_{1}$")
        axes.plot(t, np.rad2deg(phi1_analytic),
                  c="red",
                  linestyle="--",
                  linewidth=0.8,
                  label=r"$\varphi_{1,\mathrm{ana}}$")
        axes.set_xlim(left=t[0], right=t[-1])
        axes.set_xlabel(r"Time in s")
        axes.set_ylabel(r"Angle in $^{\circ}$")
        axes.grid()
        axes.legend()
        axes.legend(loc=0)

        axes1 = fig.add_subplot(212)
        axes1.plot(t, np.rad2deg(phi1_vel),
                   c="k",
                   linestyle="-",
                   label=r"$\dot{\varphi}_{1}$")
        axes1.plot(t, np.rad2deg(phi1_vel_analytic),
                   c="red",
                   linestyle="--",
                   linewidth=0.8,
                   label=r"$\dot{\varphi}_{1,\mathrm{ana}}$")
        axes1.set_xlim(left=t[0], right=t[-1])
        axes1.set_xlabel(r"Time in s")
        axes1.set_ylabel(r"Angular velocity in $^\circ\hspace{-3pt}/\mathrm{s}$")
        axes1.grid()
        axes1.legend()
        axes1.legend(loc=0)
        canvas = FigureCanvas(fig)

        plot_name = "_".join([data["regime name"], self.name, "phi1"])
        return_list.append({"name": plot_name, "figure": canvas})

        # save file
        self.write_output_files(result_name="_".join([data["regime name"], "phi1"]),
                                figure=fig,
                                output=pm.construct_result_dict(data, output={}))

        return return_list

        # parameter of the oscillation differential equation
        # D = d1*np.sqrt(g/l1)/(2*m1*l1**2)
        # omega0 = np.sqrt(l1/g)

pm.register_processing_module(pm.PostProcessingModule, TwoPendulum)
pm.register_processing_module(pm.PostProcessingModule, SimulationVerification)



# -*- coding: utf-8 -*-
import matplotlib.pyplot as plt
import numpy as np
import sympy as sp


# rcParams['text.usetex'] = True
# rcParams['text.latex.unicode'] = True


def preview(expr, **kwargs):
    """
    support function to display nice formula
    :param expr:
    :param kwargs:
    :return:
    """
    latex_str = sp.latex(expr, **kwargs)
    latex_str = latex_str.replace("operator_name", "mathrm")
    plt.text(0.1, 0.1, latex_str, fontsize=20)
    plt.axis('off')
    plt.show()

mat2array = [{'ImmutableMatrix': np.array}, 'numpy']  # settings for lambdify function

params = sp.symbols('x0_d0, x0_d1,'
                    ' phi1_d0, phi1_d1,'
                    ' phi2_d0, phi2_d1,'
                    ' m0, m1, m2, l1, l2, F, M1, M2, g, d0, d1, d2')

x0_d0, x0_d1, phi1_d0, phi1_d1, phi2_d0, phi2_d1,\
m0, m1, m2, l1, l2, F, M1, M2, g, d0, d1, d2 = params

term0 = m0 + m1*sp.sin(phi1_d0)**2 + m2*(sp.sin(phi2_d0))**2
term1 = m1*sp.sin(phi1_d0)*(g*sp.cos(phi1_d0) - l1*phi1_d1**2)
term2 = m2*sp.sin(phi2_d0)*(g*sp.cos(phi2_d0) - l2*phi2_d1**2)

x0_d2 = (term1 +
         term2 +
         (F - d0*x0_d1) +
         (M1 - d1*phi1_d1)*sp.cos(phi1_d0)/l1 +
         (M2 - d2*phi2_d1)*sp.cos(phi2_d0)/l2)/term0

phi1_d2 = g*sp.sin(phi1_d0)/l1 \
          + sp.cos(phi1_d0)*(term1 +
                             term2 +
                             (F - d0*x0_d1) +
                             (M1 - d1*phi1_d1)*sp.cos(phi1_d0)/l1 +
                             (M2 - d2*phi2_d1)*sp.cos(phi2_d0)/l2)/(l1*term0) \
          + (M1 - d1*phi1_d1)/(m1*l1**2)

phi2_d2 = g*sp.sin(phi2_d0)/l2 \
          + sp.cos(phi2_d0)*\
            (term1
             + term2
             + (F - d0*x0_d1)
             + (M1 - d1*phi1_d1)*sp.cos(phi1_d0)/l1
             + (M2 - d2*phi2_d1)*sp.cos(phi2_d0)/l2)\
            /(l2*term0) \
          + (M2 - d2*phi2_d1)/(m2*l2**2)

x = sp.Matrix([x0_d0, x0_d1, phi1_d0, phi1_d1, phi2_d0, phi2_d1])
# x = sp.Matrix([x0_d0, phi1_d0, phi2_d0, x0_d1, phi1_d1, phi2_d1])  # only for comparison
u = sp.Matrix([F, M1, M2])

sys = sp.Matrix([x0_d1, x0_d2, phi1_d1, phi1_d2, phi2_d1, phi2_d2])
# sys = sp.Matrix([x0_d1, phi1_d1, phi2_d1, x0_d2, phi1_d2, phi2_d2])  # only for comparison

dict_names = {x0_d0: r'x_{0}',
              x0_d1: r'\dot{x}_{0}',
              phi1_d0: r'\varphi_{1}',
              phi1_d1: r'\dot{\varphi}_{1}',
              phi2_d0: r'\varphi_{2}',
              phi2_d1: r'\dot{\varphi}_{2}'}
simplification_list = [(phi1_d1, 0), (phi2_d1, 0),
                       (F, 0), (M1, 0), (M2, 0),
                       # (sp.sin(phi1_d0), 0), (sp.sin(phi2_d0), 0),
                       # (sp.cos(phi1_d0)**2, 1), (sp.cos(phi2_d0)**2, 1),
                       # (sp.cos(phi1_d0)**3, sp.cos(phi1_d0)),
                       # (sp.cos(phi2_d0)**3, sp.cos(phi2_d0))
                       ]

A = sys.jacobian(x)
A = A.subs(simplification_list)
# preview(A, mode='equation', mat_str="array", symbol_names=dict_names)

# B with all control action (F, M1, M2)
B = sys.jacobian(u)
B = B.subs(simplification_list)
# preview(B, mode='equation', mat_str="array", symbol_names=dict_names)

# B with only F as control action
B = B[:, 0]

C = np.array([[1, 0, 0, 0, 0, 0]])

eq_state = [x0_d0, x0_d1, phi1_d0, phi1_d1,phi2_d0, phi2_d1]
parameter = [m0, m1, m2, l1, l2, g, d0, d1, d2]
A_func = sp.lambdify((eq_state, parameter), A, modules=mat2array)
B_func = sp.lambdify((eq_state, parameter), B, modules=mat2array)

# _eq_state = [0, 0, np.pi, 0, np.pi, 0]
# _parameter = [st.m0, st.m1, st.m2, st.l1, st.l2, st.g, st.d0, st.d1, st.d2]
#
# A_num = A_func(_eq_state, _parameter)
# B_num = B_func(_eq_state, _parameter)
#
# K = to.ackerSISO(A_num, B_num, poles=[-10.1,-8.2,-6.9,-5,-2.5,-1.5])


params = sp.symbols('x0_d0, x0_d1,'
                    ' phi1_d0, phi1_d1,'
                    ' phi2_d0, phi2_d1,'
                    'u,'
                    ' m0, m1, m2, l1, l2, F, M1, M2, g, d0, d1, d2')

x0_d0, x0_d1, phi1_d0, phi1_d1, phi2_d0, phi2_d1, u,\
m0, m1, m2, l1, l2, F, M1, M2, g, d0, d1, d2 = params

x0_d2 = u

phi1_d2 = (g*sp.sin(phi1_d0))/l1 - (d1*phi1_d1)/(m1*l1**2) + (sp.cos(phi1_d0)*u)/l1

phi2_d2 = (g*sp.sin(phi2_d0))/l2 - (d2*phi2_d1)/(m2*l2**2) + (sp.cos(phi2_d0)*u)/l2

x_vec = sp.Matrix([x0_d0, x0_d1, phi1_d0, phi1_d1, phi2_d0, phi2_d1])
# x = sp.Matrix([x0_d0, phi1_d0, phi2_d0, x0_d1, phi1_d1, phi2_d1])  # only for comparison
u_vec = sp.Matrix([u])

sys = sp.Matrix([x0_d1, x0_d2, phi1_d1, phi1_d2, phi2_d1, phi2_d2])
# sys = sp.Matrix([x0_d1, phi1_d1, phi2_d1, x0_d2, phi1_d2, phi2_d2])  # only for comparison

dict_names = {x0_d0: r'x_{0}',
              x0_d1: r'\dot{x}_{0}',
              phi1_d0: r'\varphi_{1}',
              phi1_d1: r'\dot{\varphi}_{1}',
              phi2_d0: r'\varphi_{2}',
              phi2_d1: r'\dot{\varphi}_{2}'}
simplification_list = [(phi1_d1, 0), (phi2_d1, 0),
                       (F, 0), (M1, 0), (M2, 0), (u, 0)]

A_par_lin = sys.jacobian(x_vec)
A_par_lin = A_par_lin.subs(simplification_list)
# preview(A, mode='equation', mat_str="array", symbol_names=dict_names)

B_par_lin = sys.jacobian(u_vec)
B_par_lin = B_par_lin.subs(simplification_list)
# preview(B, mode='equation', mat_str="array", symbol_names=dict_names)

C_par_lin = np.array([[1, 0, 0, 0, 0, 0]])

eq_state = [x0_d0, x0_d1, phi1_d0, phi1_d1, phi2_d0, phi2_d1]
parameter = [m0, m1, m2, l1, l2, g, d0, d1, d2]
A_func_par_lin = sp.lambdify((eq_state, parameter), A_par_lin, modules=mat2array)
B_func_par_lin = sp.lambdify((eq_state, parameter), B_par_lin, modules=mat2array)

# _eq_state = [0, 0, np.pi, 0, np.pi, 0]
# _parameter = [st.m0, st.m1, st.m2, st.l1, st.l2, st.g, st.d0, st.d1, st.d2]
#
# A_num = A_func(_eq_state, _parameter)
# B_num = B_func(_eq_state, _parameter)
#
# K = to.ackerSISO(A_num, B_num, poles=[-10.1,-8.2,-6.9,-5,-2.5,-1.5])


